#!/usr/local/bin/perl

########################################################################
# metaf.pl 1.58
#   (CGI) script to download and display METAR/TAF/SYNOP/BUOY/AMDAR messages
#
# copyright (c) 2006-2015 metaf2xml
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
# <http://www.gnu.org/licenses/>.
########################################################################

########################################################################
# some things strictly Perl
########################################################################
use strict;
use warnings;
use 5.010_001;

=head1 NAME

metaf.pl - download and display METAR/TAF/SYNOP/BUOY/AMDAR messages

=head1 SYNOPSIS

as a standalone script used from the command line:

 metaf.pl 'type_metaf=...&type_synop=...&type_buoy=...&type_amdar=...&lang=...&format=...&src_metaf=...&src_synop=...&src_buoy=...&src_amdar=...&mode=...&hours=...&msg_metaf=...&msg_synop=...&msg_buoy=...&msg_amdar=...'

or

 metaf.pl type_metaf=... type_synop=... type_buoy=... type_amdar=... lang=... format=... src_metaf=... src_synop=... src_buoy=... src_amdar=... mode=... hours=... msg_metaf='...' msg_synop='...' msg_buoy='...' msg_amdar='...'

or as a CGI script used by an HTML form (HTTP method GET):

 <form method="GET" action="metaf.pl">
  <input name="type_metaf" ... >
  <input name="type_synop" ... >
  <input name="type_buoy" ... >
  <input name="type_amdar" ... >
  <input name="lang" ... >
  <input name="format" ... >
  <input name="src_metaf" ... >
  <input name="src_synop" ... >
  <input name="src_buoy" ... >
  <input name="src_amdar" ... >
  <input name="mode" ... >
  <input name="hours" ... >
  <input name="msg_metaf" ... >
  <input name="msg_synop" ... >
  <input name="msg_buoy" ... >
  <input name="msg_amdar" ... >
 </form>

=head1 DESCRIPTION

This script can download and display in plain language up-to-date or historical
weather information (METAR, TAF, SYNOP, BUOY or AMDAR messages) from Internet
servers, an SQL database, or a local file.
It can be used from the command line, or with GET or POST requests if it is
installed to be called via CGI by a web server.

If the script is called from a web server and the file $CGI_LOG_FILE is
writable by the user running the web server (e.g. "wwwrun"), metaf.pl will log
the requests there.

=head1 DEPENDENCIES

The Perl modules C<CGI>, C<metaf2xml::parser> and C<metaf2xml::src2raw>
are required.

The Perl module C<DBI> and the module C<DBD> (for the configured type of data
source) are required if an SQL database is used as the data source.

The program C<curl> is required to download weather messages from Internet
servers.

The program C<xsltproc> is required if the script is to convert the XML to HMTL
or text.

=head1 ABBREVIATIONS

=over

=item noaa

National Oceanic and Atmospheric Administration

=item nws

the National Weather Service department of the NOAA

=item iws

the Internet Weather Source of the NWS

=item adds

the Aviation Digital Data Service of the NWS

=item addsds

the experimental data server of the Aviation Digital Data Service of the NWS

=item cod

College of DuPage

=item fsu

Florida State University

=back

=head1 OPTIONS

=over

=item -v

print version of metaf.pl and exit

=back

=head1 ARGUMENTS

The following variables can be set from CGI parameters or on the command line:

=over

=item type_metaf

type of content in C<msg_metaf>: B<metar> (default), B<taf>, B<icao> (ICAO id(s)
to fetch)

=item type_synop

type of content in C<msg_synop>: B<synop> (default), B<wmo> (WMO id(s) to
fetch), B<ship> (SYNOP SHIP id(s) to fetch), B<mobil> (SYNOP MOBIL id(s) to
fetch)

=item type_buoy

type of content in C<msg_buoy>: B<buoy> (default), B<wmo> (WMO id(s) to fetch)

=item type_amdar

type of content in C<msg_amdar>: B<amdar> (default), B<ac> (aircraft id(s) to
fetch)

=item lang

language for output of script and display of parsed message: B<de> (default),
B<en>, B<es>, B<ru>

=item format

output format: B<html> (default), B<text>, or B<xml> (with style sheet
reference)

=item src_metaf

source for METAR/TAF messages to be fetched:
B<noaa> (NOAA/IWS)[1] (default), B<adds>[2], B<addsds>[2], B<ogimet>[2],
B<nws>[1], B<cod>[1] (no TAF), B<fsu>[1] (no TAF).
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/metaf.txt>[4] exists.

Note: B<cod> and B<fsu> are not available in the web-based interfaces.

=item src_synop

source for SYNOP messages to be fetched:
B<nws>[1] (default), B<ogimet>[2], B<cod>[1], B<fsu>[1].
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/synop.txt>[4] exists.

Note: B<fsu> is not available in the web-based interfaces.

=item src_buoy

source for BUOY messages to be fetched:
B<nws>[1] (default), B<ogimet>[2].
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/buoy.txt>[4] exists.

=item src_amdar

source for AMDAR messages to be fetched:
B<nws>[1] (default), B<fsu>[1].
It is ignored if a source in C<DB_DATA_SOURCES>[3] could be opened or
F<$CGI_DATA_DIR/amdar.txt>[4] exists.

=item mode

which messages to fetch and how to display: B<latest>/detailed (default),
all/B<summary>

=item hours

for mode B<summary>: number of hours to fetch messages for (0..72, default: 24)

Note: The exact time range for which data is received depends on the source of
messages.

=item msg_metaf

METAR/TAF message to be processed, or ICAO id(s)

=item msg_synop

SYNOP message to be processed, or WMO id(s)

=item msg_buoy

BUOY message to be processed, or WMO id(s)

=item msg_amdar

AMDAR message to be processed, or aircraft id(s)

=back

[1]: Files will be fetched from the server and searched.
Note: The parameter B<hours> is used to select files by their time stamp or
name; the selected files do not necessarily contain only messages for the
choosen B<hours>.

[2]: the HTML form provided by the server will be used

[3]: For the description of the configuration variable $DB_DATA_SOURCES see the
I<USAGE NOTES> for C<DB_DATA_SOURCES> below.

[4]: For the format required for the files in F<$CGI_DATA_DIR> see the I<USAGE
NOTES> for C<CGI_DATA_DIR> below.

=cut

# all times are UTC (esp. required for mktime())
$ENV{TZ} = 'UTC';

use Getopt::Std;
use Pod::Usage 'pod2usage';
# use File::Basename 'dirname';
use POSIX qw(mktime WIFEXITED WEXITSTATUS WTERMSIG);
use CGI qw(:standard -no_xhtml *table);

our $VERSION = '1.58';

# path will be set by install
use lib '/usr/local/lib/perl5/site_perl'; # METAF2XML_LIB
use metaf2xml::parser 1.058 qw(start_xml parse_report finish);
use metaf2xml::src2raw 1.058 qw(start_file next_line);

#
# restrict use of CGI:
#

# uploads are not allowed
$CGI::DISABLE_UPLOADS = 1;

=head1 USAGE NOTES

The length of the body of the POST request (not the length of query string of
the URL) is restricted to 600 bytes.

=cut

$CGI::POST_MAX = 600;

########################################################################
# config section, possibly changed by install
########################################################################

=pod

The following variables can be adapted to local needs; this is typically done
during the installation:

=over

=item METAF2XML_XSL

directory for the XSL style sheets

=cut

# path will be set by install
my $METAF2XML_XSL = '/usr/local/share/metaf2xml';

=item CGI_DATA_DIR

directory for files with pre-processed METAR/TAF/SYNOP/BUOY/AMDAR messages

The messages in these files must have the format specified by the I<WMO Manual
No. 306>, without modifications due to distribution like providing the initial
part of messages only once for several messages or appending an "=" (equal sign)
to terminate a message, with the report type C<TT> prepended. This format can
be obtained using:

 metafsrc2raw.pl -F<FORMAT> -H1 [FILE] ...

=cut

my $CGI_DATA_DIR = '/var/tmp';

=item CGI_LOG_FILE

complete path for log file

=cut

my $CGI_LOG_FILE = '/var/tmp/metaf2xml.log';

=item CURL

path and options for C<curl>

=cut

my $CURL = 'curl -sSf --stderr -';

=item CURL_PROXY_OPTS

options for C<curl> if an HTTP proxy should be used

=cut

my $CURL_PROXY_OPTS = '';

=item DB_DATA_SOURCES

space delimited strings to describe possible data sources

If the string C<sqlite> is in the list, the file F<$CGI_DATA_DIR/sqlite.db> is
used as database.
For other strings, the file F<$CGI_DATA_DIR/>C<string>F<.db> should contain the
connect string (C<DBI:>...) in the first line. The database username is
C<metaf2xml>, it has no password; the user C<metaf2xml> should have only SELECT
rights.

The first data source which can be opened is used. If the last data source fails
to open or C<DB_DATA_SOURCES> is empty, the parameters C<src_>* are used.

However, if a string ends in C</f>, this is the last data source to try and the
parameters C<src_>* are not used, either.
If this source fails to open, the script will exit with the HTTP error 503.

The Perl module C<DBI> and the module C<DBD> for the configured data source are
required if a database is to be used.

=cut

my $DB_DATA_SOURCES = '';

=back

=cut

#
# define some regular expressions
#
my $re_report_type_synop = '(?:S[IMN])';
my $re_report_type_metaf = '(?:S[AP]|F[CT])';
my $re_header_synop = '(?:AAXX .{5}|(?:BB|OO)XX)';
my $re_header_buoy = '(?:ZZYY)';
my $re_header_amdar = '(?:AMDAR \d{4} (?:[A-Z]{3}|///))';
my $re_ICAO = '(?:[A-Z][A-Z\d]{3})';
my $re_day = '(?:[0-2]\d|3[01])';
my $re_hour = '(?:[01]\d|2[0-3])';
my $re_min  = '[0-5]\d';
my $re_header_metaf = '(?:COR |AMD |METAR |TAF |SPECI )*';
my $re_header_sao = "([A-Z]{3,4}) (S[AP]|RS)(?: COR)? $re_hour$re_min (?:AUTO|AWOS|SAWR)";
my $re_aircraft = '[A-Z]{2}[A-Z0-9]{2,6}';

#
# make a list of strings unique
#
sub mk_uniq {
    my $uniq = '';
    for (@_) {
        if ($uniq eq '') {
            $uniq = $_;
        } elsif (index(" $uniq ", " $_ ") == -1) {
            $uniq .= " $_";
        }
    }
    return $uniq;
}

#
# provide info for processes with errors
#
sub exit_info {
    my ($error, $rc) = @_;
    my $wifexited;

    $wifexited = eval { WIFEXITED($rc) };

    return   'ERROR - '
           . ($error ? "$error - " : '')
           . (defined $wifexited
                             ? ($wifexited ? 'exit(' . WEXITSTATUS($rc) . ')'
                                           : 'signal(' . WTERMSIG($rc) . ')')
                             : "retval($rc)")
           . ' - ';
}

#
# provide pipe open depending on OS
#
sub pipe_open {
    my ($fh, $is_read, @args) = @_;

    if ({ map { $_ => 1 } qw(dos MSWin32 NetWare MacOS riscos) }->{$^O}) {
        my ($at_beginning, $at_end);

        ($at_beginning, $at_end) = $is_read ? ('','|') : ('|', '');

        # quote spaces if necessary
        return open $$fh, $at_beginning . join(' ', map { / / ? "\"$_\"" : $_ } @args) . $at_end;
    }
    return open $$fh, $is_read ? '-|' : '|-', @args;
}

#
# determine max. available hours
#
sub max_hours {
    my ($max_hours, $hours, $end_date, $round6h) = @_;

    $end_date = time
        unless defined $end_date;

    if ($max_hours > 0) {
        my ($hh_end_date, $hh_start_avail, $hh_avail);

        $hh_end_date    = $end_date;
        $hh_start_avail = time;
        $hh_start_avail -= $hh_start_avail % 3600;
        $hh_start_avail += (6 - (($hh_start_avail / 3600) % 6)) * 3600
            if $round6h;
        $hh_start_avail -= $max_hours * 3600;
        $hh_end_date    -= $hh_end_date % 3600;
        $hh_start_avail -= $hh_start_avail % 3600;
        $hh_avail       = (  $hh_end_date - $hh_end_date % 3600
                           - $hh_start_avail) / 3600;
        $hours = $hh_avail
            if $hours > $hh_avail;
    }

    return ($hours, $end_date);
}

#
# get the contents of the directory on the NOAA/NWS server from the server
#   type: metar, taf, land, ship, buoy, amdar
sub files_nws {
    my ($type, $hours) = @_;
    my ($url, $cmd, $fh, $response, @files);

    # fetch directory listing
    # also:      ftp://tgftp.nws.noaa.gov/SL.us008001/DF.an/
    $url =   'http://weather.noaa.gov/pub/SL.us008001/DF.an/'
           . { metar => 'DC.sflnd/DS.metar',
               taf   => 'DC.sflnd/DS.tafst',
               land  => 'DC.sflnd/DS.synop',
               ship  => 'DC.sfmar/DS.ships',
               buoy  => 'DC.sfmar/DS.dbuoy',
               amdar => 'DC.sluan/DS.amdar',
             }->{$type};
    $cmd = "$CURL $CURL_PROXY_OPTS $url/ls-lt";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"), "$url/ls-lt"
        or return "ERROR - $cmd - $!";

    {
        local $/;
        $response = <$fh>;
    }
    if (!close $fh || $?) {
        $response =~ s/[\n\r]+/ /g;
        return exit_info($!, $?) . "$cmd - $response";
    }

    @files = $response =~ / (sn.\d{4}.txt)/g;

    # don't use the oldest file, it may have/may shortly become the newest
    $#files--;

    # truncate array if less hours are requested
    # (one file every 5 min. = 12 per hour)
    $#files = $hours * 12
        if $hours * 12 < $#files;

    return $url, @files;
}

#
# determine file names for data from NOAA/IWS (cycle) files
#   type: metar, staf, taf
sub files_cycle_iws {
    my ($type, $hours, $end_date) = @_;
    my ($hour, @hours);

    ($hours, $end_date) = max_hours 24, $hours, $end_date, $type ne 'metar';

    if ($type eq 'metar') {
        # METAR cycle files start 15 min. before the hour
        $hour = (($end_date + 15 * 60) / 3600) % 24;
        @hours = map { ($hour - $_) % 24 } 0 .. $hours - 1;
    } else {
        $hour = ($end_date / 3600) % 24;
        for (0 .. $hours - 1) {
            my $h;

            $h = $hour - $_;
            $h = ($h - $h % 6) % 24;
            push @hours, $h
                if $#hours == -1 || $hours[-1] != $h;
        }
    }
    return   'http://weather.noaa.gov/pub/data/'
           . { metar => 'observations/metar',
               staf  => 'forecasts/shorttaf',
               taf   => 'forecasts/taf',
             }->{$type}
           . '/cycles',
           map { sprintf '%02dZ.TXT', $_ } @hours;
}

#
# determine file names for hourly files
#
sub files_hourly {
    my ($max_hours, $hours, $end_date, $format, @ids) = @_;
    my @files;

    ($hours, $end_date) = max_hours $max_hours, $hours, $end_date;

    # one file every hour
    for (0 .. $hours - 1) {
        my @date_arr;

        @date_arr = (gmtime $end_date - $_ * 60 * 60)[5, 5, 4, 3, 2];
        $date_arr[0] %= 100;  # year is offset to 1900, need 2 digits, only
        $date_arr[1] += 1900; # year is offset to 1900, need 4 digits
        $date_arr[2]++;       # month starts from 0
        push @files, sprintf $format, @date_arr[@ids];
    }
    return @files;
}

#
# determine file names for data from College of DuPage
#   type: syn, sao
sub files_cod {
    my ($type, $hours, $end_date) = @_;

    return "http://weather.cod.edu/digatmos/$type",
           files_hourly 60, $hours, $end_date,
                        "%02d%02d%02d%02d.$type", 0, 2 .. 4;
}

#
# determine file names for data from Florida State University
#   type: syn, metar, aircraft/acrft
sub files_fsu {
    my ($type, $hours, $end_date, $suff) = @_;

    $suff //= $type;
    return "http://www.eoas.fsu.edu/rawdata/$type",
           files_hourly 168, $hours, $end_date,
                        "%d%02d%02d/%d%02d%02d%02d.$suff", 1 .. 3, 1 .. 4;
}

#
# determine file names for data from Iowa State University
#  - historical text files since 2005-07-25 (except for 2011 .. 2013)
#  - provided at 08:10 for the previous day
#  - METAR: selected stations, only
#  - TAF: none
#  - SYNOP: not all stations for all observation times, but some with nat. data
#
#   type: syn, boy, sao, upa
sub files_isu {
    my ($type, $hours, $end_date) = @_;

    # also:  ftp://mtarchive.geol.iastate.edu/data
    return 'http://mtarchive.geol.iastate.edu',
           files_hourly -1, $hours, $end_date,
                        "%d/%02d/%02d/text/$type/%d%02d%02d%02d_$type.txt",
                        1 .. 3, 1 .. 4;
}

#
# handle data from opened files
#
{
    my @latest;

    sub start_fh {
        @latest = ();
        return;
    }

    sub next_fh {
        my $fh = shift;

        # line(s) from previous call?
        return pop @latest
            if $#latest > -1;

        while (1) {
            # get one line from input and parse
            $_ = <$fh>;
            @latest = next_line $_;

            # return EOF on error
            last
                if $#latest == 0 && !defined $latest[0];
            # have line(s)? return first one, others ones are for next call(s)
            return pop @latest
                if $#latest > -1;
            # found no lines and input has EOF? return EOF
            last
                unless defined $_;
        }
        return;
    }
}

#
# scan files with SYNOP data for $wmoid
#
sub search_files_synop {
    my ($wmoid, $mode, $src, $hours, $type_synop, $end_date) = @_;
    my ($url, @files, @wmoids, $no_wmoid, @arr);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/synop.txt")
                     : $src eq 'cod'   ? files_cod('syn', $hours, $end_date)
                     : $src eq 'fsu'   ? files_fsu('syn', $hours, $end_date)
                     : $src eq 'isu'   ? files_isu('syn', $hours, $end_date)
                     : $src eq 'nws'   ? files_nws($type_synop eq 'ship'
                                                      ? 'ship' : 'land', $hours)
                     :                   ();
    return ()
        unless defined $url;
    return $url
        if $src eq 'nws' && $url =~ /ERROR/;

    $wmoid =~ s/ /|/g;
    @wmoids = split /[|]/, $wmoid;
    $no_wmoid = $#wmoids + 1;
    @arr = ();
    for my $file (@files) {
        my ($cmd, $fh, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("synop_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return "ERROR - $wmoid - $cmd - $!";
        } else {
            $cmd = $file;
            open $fh, '<', $file
                or return "ERROR - $wmoid - $cmd - $!";
        }

        @file_arr = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $space, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            ($report_type, $space, $msg) = unpack 'a2aa*', $line;
            next unless $space eq ' ';

            ($id) = $msg =~ m{^$re_header_synop ([A-Z\d/]{3,})}o;
            next unless defined $id;

            # include station in $line
            $line = "$report_type $id $msg";

            $report_type_id = "$report_type-$id";
            $report_type_matches = index("|$wmoid|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$wmoid|", "|$id|") > -1;

            # looking for messages with this id?
            if ($report_type_matches) {
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                $report_type_allowed = $re_report_type_synop;
            } else {
                next;
            }

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @file_arr, $line
                    unless grep { /^\Q$line\E$/ } (@arr, @file_arr);
                next;
            }

            # found message in previous file(s)?
            next if grep { /^$report_type_allowed $re_header_synop $id / } @arr;

            # request  : input             => output
            # id       : Sy-id,Sx-id,Sz-id => Sz-id
            # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
            # Sx-id,id : Sx-id,Sy-id       => Sx-id,Sy-id
            # Sx-id,id : Sy-id,Sx-id       => Sx-id
            # looking for messages with the report type of this message?
            if ($report_type_matches) {
                if ($msg_found{$report_type_id}) {  # record type found already?
                    $msg_seq =~ s/ $report_type_id //;   # remove existing entry
                } else {
                    $no_wmoid--;                    # mark record type as found
                }
                $msg_seq = " $report_type_id " . $msg_seq;   # prepend new entry
                $msg_found{$report_type_id} = $line;         # save new value
            }

            # looking for messages with any report type?
            if ($report_type_any) {
                if ($msg_found{"Sx-$id"}) {
                    $msg_seq =~ s/ Sx-$id //;
                } else {
                    $no_wmoid--;
                }
                # prepend only if not done already above
                $msg_seq = " Sx-$id " . $msg_seq
                    unless $report_type_matches;
                $msg_found{"Sx-$id"} = $line;
            }
        }
        return exit_info($!, $?) . "$wmoid - $cmd"
            if !close $fh || $?;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_wmoid == 0;
    }
    return @arr;
}

#
# scan files with BUOY data for $wmoid
#
sub search_files_buoy {
    my ($wmoid, $mode, $src, $hours, $end_date) = @_;
    my ($url, @files, @wmoids, $no_wmoid, @arr);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/buoy.txt")
                     : $src eq 'isu'   ? files_isu('boy',  $hours, $end_date)
                     : $src eq 'nws'   ? files_nws('buoy', $hours)
                     :                   ();
    return ()
        unless defined $url;
    return $url
        if $src eq 'nws' && $url =~ /ERROR/;

    $wmoid =~ s/ /|/g;
    @wmoids = split /[|]/, $wmoid;
    $no_wmoid = $#wmoids + 1;
    @arr = ();
    for my $file (@files) {
        my ($cmd, $fh, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("buoy_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return "ERROR - $wmoid - $cmd - $!";
        } else {
            $cmd = $file;
            open $fh, '<', $file
                or return "ERROR - $wmoid - $cmd - $!";
        }

        @file_arr = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $space, $msg, $id);

            ($report_type, $space, $msg) = unpack 'a2aa*', $line;
            next unless $space eq ' ';

            ($id) = $msg =~ m{^$re_header_buoy (\d{5})}o;
            next unless defined $id && index("|$wmoid|", "|$id|") > -1;

            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @file_arr, $line
                    unless grep { /^\Q$line\E$/ } (@arr, @file_arr);
                next;
            }

            # found message in previous file(s)?
            next if grep { /^$report_type $re_header_buoy $id / } @arr;

            if ($msg_found{$id}) {
                $msg_seq =~ s/ $id //;
            } else {
                $no_wmoid--;
            }
            $msg_seq = " $id " . $msg_seq;   # prepend new entry
            $msg_found{$id} = $line;

        }
        return exit_info($!, $?) . "$wmoid - $cmd"
            if !close $fh || $?;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_wmoid == 0;
    }

    return @arr;
}

#
# scan files with AMDAR data for $acid
#
sub search_files_amdar {
    my ($acid, $mode, $src, $hours, $end_date) = @_;
    my ($url, @files, @acids, $no_acid, @arr);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local' ? ('', "$CGI_DATA_DIR/amdar.txt")
                     : $src eq 'fsu'   ?
                               files_fsu('aircraft', $hours, $end_date, 'acrft')
                     : $src eq 'isu'   ? files_isu('upa',  $hours, $end_date)
                     : $src eq 'nws'   ? files_nws('amdar', $hours)
                     :                   ();
    return ()
        unless defined $url;
    return $url
        if $src eq 'nws' && $url =~ /ERROR/;

    $acid =~ s/ /|/g;
    @acids = split /[|]/, $acid;
    $no_acid = $#acids + 1;
    @arr = ();
    for my $file (@files) {
        my ($cmd, $fh, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file("amdar_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return "ERROR - $acid - $cmd - $!";
        } else {
            $cmd = $file;
            open $fh, '<', $file
                or return "ERROR - $acid - $cmd - $!";
        }

        @file_arr = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $space, $msg, $id, $msg_proper);

            ($report_type, $space, $msg) = unpack 'a2aa*', $line;
            next unless $space eq ' ';

            ($id) = $msg =~ m{^$re_header_amdar ($re_aircraft)}o;
            next unless defined $id && index("|$acid|", "|$id|") > -1;

            if ($mode eq 'summary') {
                # prepend unless it already exists
                # ignore "UD AMDAR YYGG ", YYGG is useless
                $msg_proper = substr $line, 14;
                for (@arr, @file_arr) {
                    if ($msg_proper eq substr $_, 14) {
                        $msg_proper = undef;
                        last;
                    }
                }
                unshift @file_arr, $line
                    if $msg_proper;
                next;
            }

            # if msg was in previous files: ignore this msg, this file is older
            $msg_proper = substr $line, 14;
            for (@arr) {
                if ($msg_proper eq substr $_, 14) {
                    $msg_proper = undef;
                    last;
                }
            }
            next
                unless $msg_proper;

            if ($msg_found{$id}) {
                $msg_seq =~ s/ $id //;
            } else {
                $no_acid--;
            }
            $msg_seq = " $id " . $msg_seq;   # prepend new entry
            $msg_found{$id} = $line;

        }
        return exit_info($!, $?) . "$acid - $cmd"
            if !close $fh || $?;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_acid == 0;
    }

    return @arr;
}

#
# scan files with METAR/TAF data for $icao
#
sub search_files_metaf {
    my ($icao, $mode, $src, $hours, $type_metaf, $end_date) = @_;
    my ($url, @files, @icao, $no_icao, @arr);

    return ()
        if $icao eq '' || $icao eq ' ' || $icao eq '  ';

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    ($url, @files) =   $src eq 'local'     ? ('', "$CGI_DATA_DIR/metaf.txt")
                     : $src eq 'cod'       ? files_cod('sao',  $hours,$end_date)
                     : $src eq 'fsu'       ? files_fsu('metar',$hours,$end_date)
                     : $src eq 'isu'       ? files_isu('sao',  $hours,$end_date)
                     : $src eq 'nws'       ? files_nws($type_metaf, $hours)
                     : $src eq 'cycle_iws' ? files_cycle_iws($type_metaf,
                                                             $hours, $end_date)
                     :                       ();
    return ()
        unless defined $url;
    return $url
        if $src eq 'nws' && $url =~ /ERROR/;

    $icao =~ s/ /|/g;
    @icao = split /[|]/, $icao;
    $no_icao = @icao;
    @arr = ();
    for my $file (@files) {
        my ($cmd, $fh, @file_arr, %msg_found, $msg_seq);

        if ($url) {
            $cmd = "$CURL $CURL_PROXY_OPTS $url/$file";
            start_file(($src eq 'cycle_iws' ? $type_metaf : 'metaf') . "_$src", 1, 1);
            start_fh;
            pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS"),"$url/$file"
                or return "ERROR - $icao - $cmd - $!";
        } else {
            $cmd = $file;
            open $fh, '<', $file
                or return "ERROR - $icao - $cmd - $!";
        }

        @file_arr = ();
        $msg_seq  = '';
        while (my $line = $url ? next_fh $fh : <$fh>) {
            my ($report_type, $space, $report_type_allowed, $msg, $id);
            my ($report_type_id, $report_type_matches, $report_type_any);

            return $line if $line =~ /ERROR/;

            ($report_type, $space, $msg) = unpack 'a2aa*', $line;
            next unless $space eq ' ';

            ($id) = $msg =~ m{^$re_header_metaf($re_ICAO) }o;
            next unless defined $id;

            $report_type_id = "$report_type-$id";
            $report_type_matches = index("|$icao|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$icao|", "|$id|") > -1;

            # looking for messages with this id?
            if ($report_type_matches) {
                $report_type_allowed = $report_type;
            } elsif ($report_type_any) {
                $report_type_allowed = $re_report_type_metaf;
            } else {
                next;
            }

            # looking for all messages (with correct id)?
            if ($mode eq 'summary') {
                # prepend unless it already exists
                unshift @file_arr, $line
                    unless grep { /^\Q$line\E$/ } (@arr, @file_arr);
                next;
            }

            # found message in previous file(s)?
            next if grep { /^$report_type_allowed $re_header_metaf$id / } @arr;

            # looking for messages with the report type of this message?
            if ($report_type_matches) {
                if ($msg_found{$report_type_id}) {  # record type found already?
                    $msg_seq =~ s/ $report_type_id //;   # remove existing entry
                    $msg_seq =~ s/ SP-$id //             # SA: also remove SP
                        if $report_type eq 'SA';
                } else {
                    $no_icao--;                   # mark record type as found
                    $no_icao--                    # SA: don't search for more SP
                        if $report_type eq 'SA' && !$msg_found{"SP-$id"};
                }
                $msg_seq = " $report_type_id " . $msg_seq;   # prepend new entry
                $msg_found{$report_type_id} = $line;         # save new value
                $msg_found{"SP-$id"} = $line                 # SA: mark SP found
                    if $report_type eq 'SA';
            }
        }
        return exit_info($!, $?) . "$icao - $cmd"
            if !close $fh || $?;

        # fill file_arr if not done yet
        @file_arr = map { $msg_found{$_} } $msg_seq =~ /[^ ]+/g
            if $mode ne 'summary';

        # @arr: x+6, x+5. @file_arr: x+4, x+3.
        @arr = (@arr, @file_arr);

        last if $mode eq 'latest' && $no_icao == 0;
    }
    return @arr;
}

#
# query Ogimet
#
sub query_ogimet {
    my ($id, $mode, $which, $hours, $rep_type, $end_date) = @_;
    my (@date_inicio, @date_fin, $data, $url, $cmd, $fh, $response);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest
    $end_date = time unless defined $end_date;
    @date_inicio = (gmtime $end_date - $hours * 3600 + 60)[5, 4, 3, 2, 1];
    $date_inicio[0] += 1900;              # year is offset to 1900
    $date_inicio[1]++;                    # month starts from 0
    @date_fin = (gmtime $end_date)[5, 4, 3, 2, 1];
    $date_fin[0] += 1900;                 # year is offset to 1900
    $date_fin[1]++;                       # month starts from 0
    $data =   '-Gdtipo=' . $rep_type . ' -dord=REV -dnil=SI -dfmt=txt'
            . sprintf(  ' -dano=%d -dmes=%02d -dday=%02d -dhora=%02d -dmin=%02d'
                      . ' -danof=%d -dmesf=%02d -ddayf=%02d -dhoraf=%02d'
                      . ' -dminf=%02d',
                      @date_inicio, @date_fin)
            . ' -d' . ($which eq 'synopsc2' ? 'estado' : 'lugar') . "=$id";
    $url = 'http://www.ogimet.com/display_' . $which . '.php';
    $cmd = "$CURL $CURL_PROXY_OPTS $data $url";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS $data"), $url
        or return "ERROR - $cmd - $!";

    {
        local $/;
        $response = <$fh>;
    }
    if (!close $fh || $?) {
        $response =~ s/[\n\r]+/ /g;
        return exit_info($!, $?) . "$cmd - $response";
    }
    $response =~ s/.*(?=^.*quota limit.*reached)/ERROR: /msi;
    return $response;
}

#
# query Ogimet for $wmoid
#
sub query_synop_ogimet {
    my ($wmoid, $mode, $type_synop, $wmoid_uniq, $hours, $rep_type, $end_date) = @_;
    my ($response, @arr, %msg_found, $is_by_ctry);

    $wmoid_uniq =~ s/ /+/g;
    $is_by_ctry = $type_synop eq 'wmo' && $wmoid_uniq =~ /[A-Z]$/ ? 'c' : '';
    $response = query_ogimet $wmoid_uniq, $mode,
                ($type_synop eq 'ship' ? 'ships' : "synops${is_by_ctry}2"),
                $hours, $rep_type, $end_date;
    return $response if $response =~ /ERROR/;

    $wmoid =~ s/ /|/g;

    $_ = (split /<[\/]?pre>/, $response)[1];     # extract answer
    return "ERROR - $wmoid_uniq - no data" unless defined $_;
    s/^#.*//mg;                                  # make comment lines empty
    s/\n+/\n/g;                                  # delete consecutive newlines
    s/\n +/ /g;                                  # delete newlines in messages
    s/=//g;                                      # delete EOM
    s/^.{13}//mg;                                # remove time stamps
    for (split /\n/) {                           # process each message
        my ($id, $report_time, $report_type, $was_added);
        my ($report_type_id, $report_type_matches, $report_type_any);

        if (m{^AAXX $re_day($re_hour)[0134/] ([A-Z\d]{3,}) }o) {
            $report_time = $1;
            $id = $2;
        } elsif (m{^(?:BB|OO)XX ([A-Z\d]{3,}) $re_day($re_hour)[0134/] }o) {
            $id = $1;
            $report_time = $2;
        } else {
            next;
        }

        if ($report_time =~ /00|06|12|18/) {
            $report_type = 'SM';
        } elsif ($report_time =~ /03|09|15|21/) {
            $report_type = 'SI';
        } else {
            $report_type = 'SN';
        }

        $report_type_id = "$report_type-$id";
        if ($is_by_ctry) {
            $report_type_matches = index($wmoid, $report_type) > -1;
            $report_type_any     = $wmoid !~ /^$re_report_type_synop-/;
        } else {
            $report_type_matches = index("|$wmoid|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$wmoid|", "|$id|") > -1;
        }

        if ($mode eq 'summary') {
            push @arr, "$report_type $id $_"
                if $report_type_matches || $report_type_any;
            next;
        }

        # request  : input             => output
        # id       : Sy-id,Sx-id,Sz-id => Sy-id
        # Sx-id    : Sy-id,Sx-id,Sz-id => Sx-id
        # Sx-id,id : Sx-id,Sy-id       => Sx-id
        # Sx-id,id : Sy-id,Sx-id       => Sy-id,Sx-id
        $was_added = 0;
        if ($report_type_matches && !exists $msg_found{$report_type_id}) {
            $msg_found{$report_type_id} = undef;
            $was_added = 1;
            push @arr, "$report_type $id $_";
        }
        if ($report_type_any && !exists $msg_found{"Sx-$id"}) {
            $msg_found{"Sx-$id"} = undef;
            push @arr, "$report_type $id $_"
                unless $was_added;
        }
    }
    return @arr;
}

#
# query Ogimet for $wmoid
#
sub query_buoy_ogimet {
    my ($mode, $wmoid_uniq, $hours, $rep_type, $end_date) = @_;
    my ($response, @arr, %msg_found);

    $wmoid_uniq =~ s/ /+/g;
    $response = query_ogimet $wmoid_uniq, $mode,
                             'buoys', $hours, $rep_type, $end_date;
    return $response if $response =~ /ERROR/;

    $_ = (split /<[\/]?pre>/, $response)[1];     # extract answer
    return "ERROR - $wmoid_uniq - no data" unless defined $_;
    s/^#.*//mg;                                  # make comment lines empty
    s/\n+/\n/g;                                  # delete consecutive newlines
    s/\n +/ /g;                                  # delete newlines in messages
    s/=//g;                                      # delete EOM
    s/^.{13}//mg;                                # remove time stamps
    for (split /\n/) {                           # process each message
        /^$re_header_buoy (\d{5}) /o
            or next;
        if ($mode eq 'summary' || !exists $msg_found{$1}) {
            push @arr, 'SS ' . $_;
            $msg_found{$1} = undef;
        }
    }
    return @arr;
}

#
# query NOAA/ADDS for $icao
#
sub query_metaf_adds {
    my ($icao, $mode, $hours) = @_;
    my ($data, $url, $cmd, $fh, $response);

    $icao =~ s/ /+/g;

    $data = '-Gdstd_trans=standard -dchk_metars=on -dsubmitmet=Submit -dhoursStr=';
    if ($mode eq 'summary') {
        # only integer hours allowed
        $data .= 'past+' . int($hours) . '+hours';
    } else {
        $data .= 'most+recent+only';
    }
    $data .= " -dchk_tafs=on -dstation_ids=$icao";
    $url  = 'http://aviationweather.gov/adds/metars/';
    $cmd = "$CURL $CURL_PROXY_OPTS $data $url";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS $data"), $url
        or return "ERROR - $cmd - $!";

    {
        local $/;
        $response = <$fh>;
    }
    if (!close $fh || $?) {
        $response =~ s/[\n\r]+/ /g;
        return exit_info($!, $?) . "$cmd - $response";
    }
    $response =~ s/([^>]<font face=.mono[^>]*?>)/$1SA /ig;
    $response =~ s/([^>]<font face=.mono[^>]*?>SA SPECI )/$1SP /ig;
    $response =~ s/(<pre><font face=.mono[^>]*?>)/$1FT /ig;
    $response =~ s/[\n\r]+//g;
    $response =~ s/ +/ /g;
    $response =~ s/ +</</g;
    return $response =~ /(?:<font face=.mono[^>]*?>)([^<]+)/ig;
}

# ADDS dataserver does not provide TAF with NIL
sub query_metaf_addsds_get {
    my ($icao, $type, $hours, $end_date) = @_;
    my ($data, $url, $cmd, $fh, $response);

    return '' if $icao eq '';

    $icao =~ s/ /,/g;
    $end_date = time unless defined $end_date;

    $data = '-GdrequestType=retrieve -dformat=xml -ddataSource=';
    if ($type eq 'metar') {
        $data .= 'metars -dfields=raw_text,station_id,observation_time,metar_type';
    } else {
        $data .= 'tafs -dfields=raw_text,station_id,issue_time,valid_time_from,valid_time_to -dtimeType=issue';
    }
    $data .=   ' -dstartTime=' . ($end_date - $hours * 3600)
             . ' -dendTime=' . $end_date
             . " -dstationString=$icao";
    $url = 'http://aviationweather.gov/adds/dataserver1_3/httpparam';
    $cmd = "$CURL $CURL_PROXY_OPTS $data $url";
    pipe_open \$fh, 1, split(' ', "$CURL $CURL_PROXY_OPTS $data"), $url
        or return "ERROR - $cmd - $!";

    {
        local $/;
        $response = <$fh>;
    }
    $response =~ s/[\n\r]+/ /g;
    return exit_info($!, $?) . "$cmd - $response"
        if !close $fh || $?;
    return $response;
}

#
# query NOAA/ADDS dataserver for $icao
#
sub query_metaf_addsds {
    my ($icao_metar, $icao_taf, $mode, $hours, $end_date, $msg_metaf) = @_;
    my ($response, %msg_found, @arr, %metar, %taf);

    $hours = 7 unless $mode eq 'summary'; # overwrite for mode latest

    $response = query_metaf_addsds_get
                mk_uniq($icao_metar =~ /$re_report_type_metaf-([^ ]+)/og),
                'metar', $hours, $end_date;
    return $response if $response =~ /^ERROR/;

    $msg_metaf =~ s/ /|/g;
    $response =~ s/ +/ /g;
    %metar = ();
    %msg_found = ();
    # use closing tag for METAR, metar_type is sometimes missing
    for ($response =~ m{<((?:raw_text|station_id|observation_time|metar_type|/METAR)>[^<]+)}ig)
    {
        /^([^>]+).(.*)/;
        if ($1 ne '/METAR') {
            $metar{$1} = $2;
            next;
        }

        # check for missing fields
        $metar{metar_type} = $metar{metar_type} && $metar{metar_type} eq 'SPECI'
                             ? 'SP' : 'SA';
        if (scalar keys %metar != 4) {
            %metar = ();
            next;
        }

        if ($mode ne 'summary') {
            my $key = "$metar{station_id},$metar{metar_type}";
            if (exists $msg_found{$key}) {
                %metar = ();
                next;
            }

            $msg_found{$key} = undef;
            # SA found first and no type specified: mark SP found
            $msg_found{"$metar{station_id},SP"} = undef
                if    $metar{metar_type} eq 'SA'
                   && index("|$msg_metaf|", "|SP-$metar{station_id}|") == -1
                   && index("|$msg_metaf|", "|SA-$metar{station_id}|") == -1;
        }

        $metar{raw_text} =~ s/^(?:METAR |SPECI )*//;
        push @arr, "@metar{qw(observation_time metar_type raw_text)}";
        %metar = ();
    }

    for ($response =~ /(<error>[^<]+)/ig) {
        /^[^>]+.(.*)/;
        push @arr, ' ' x 21 . "ERROR - $icao_metar - $1";
    }

    $response = query_metaf_addsds_get
                mk_uniq($icao_taf =~ /$re_report_type_metaf-([^ ]+)/og),
                'taf', $hours, $end_date;
    return $response if $response =~ /^ERROR/;

    %taf = ();
    %msg_found = ();
    for ($response =~ /<((?:raw_text|station_id|issue_time|valid_time_from|valid_time_to)>[^<]+)/ig)
    {
        /^([^>]+).(.*)/
            or next;
        $taf{$1} = $2;
        next
            if $1 ne 'valid_time_to';

        # check for missing fields
        if (scalar keys %taf != 5) {
            %taf = ();
            next;
        }

        my @date_arr;

        @date_arr[5,4,3,2,1,0] = split /[-TZ:]/, $taf{valid_time_from};
        $date_arr[5] -= 1900;                           # year is offset to 1900
        $date_arr[4]--;                                 # month starts from 0
        $taf{valid_time_from} = mktime @date_arr, -1, -1, -1;
        @date_arr[5,4,3,2,1,0] = split /[-TZ:]/, $taf{valid_time_to};
        $date_arr[5] -= 1900;                           # year is offset to 1900
        $date_arr[4]--;                                 # month starts from 0
        $taf{valid_time_to} = mktime @date_arr, -1, -1, -1;
        $hours = ($taf{valid_time_to} - $taf{valid_time_from}) / 3600;
        $taf{msg_type} = $hours >= 12 ? 'FT' : 'FC';

        if ($mode ne 'summary') {
            my $key = "$taf{station_id},$taf{msg_type}";
            if (exists $msg_found{$key}) {
                %taf = ();
                next;
            }

            $msg_found{$key} = undef;
        }

        $taf{raw_text} =~ s/^(?:TAF )*//;
        push @arr, "@taf{qw(issue_time msg_type raw_text)}";
        %taf = ();
    }

    for ($response =~ /(<error>[^<]+)/ig) {
        /^[^>]+.(.*)/;
        push @arr, ' ' x 21 . "ERROR - $icao_metar - $1";
    }

    return map { substr $_, 21 } sort {$b cmp $a} @arr;
}

#
# query Ogimet for $icao
#
sub query_metaf_ogimet {
    my ($icao, $mode, $hours, $rep_type, $end_date, $msg_metaf) = @_;
    my ($response, @msgs, @arr, %msg_found, $key, $key1, $key2);

    $icao =~ s/ /+/g;
    # also returns TAF!
    $response = query_ogimet $icao, $mode, 'metars2', $hours, $rep_type, $end_date;
    return $response if $response =~ /ERROR/;

    $_ = (split /<[\/]?pre>/, $response)[1];  # extract answer
    return "ERROR - $icao - no data" unless defined $_;
    s/^#  (?=TAF|MET)/x            /mg;      # preserve comments w. message type
    s/^#.*//mg;                              # make comment lines empty
    s/\n+/\n/g;                              # delete consecutive newlines
    s/\n +/ /g;                              # delete newlines in messages
    s/=//g;                                  # delete EOM

    # prepend message type
    @msgs = ();
    for (split /^x/m) {
        if (/^ +METAR/) {
            s/.*\n//;
            s/^(.{12})/$1 SA/mg;
            push @msgs, split /\n/;
        } elsif (/^ +TAF LAR/) {
            s/.*\n//;
            s/^(.{12})/$1 FT/mg;
            push @msgs, split /\n/;
        } elsif (/^ +TAF COR/) {
            s/.*\n//;
            s/^(.{12})/$1 FC/mg;
            push @msgs, split /\n/;
        }
    }

    $msg_metaf =~ s/ /|/g;
    for (sort {$b cmp $a} @msgs) {              # process each message
        s/^.{13}//;                             # remove time stamps
        s/^SA SPECI /SP /;                      # remove keywords
        s/^SA METAR /SA /;
        s/^(F[CT]) TAF /$1 /;

        if ($mode ne 'summary') {
            /^($re_report_type_metaf) $re_header_metaf($re_ICAO) /o;
            if (defined $1) {
                ($key1, $key2) = ($1, $2);
            } else {
                /^$re_header_sao/o
                    or next;
                ($key1, $key2) = ($2, $1);
                if (length $key2 == 3) {
                    $key2 = "C$key2";
                    s/^/C/;
                }
                s/^/$key1 /;
            }

            $key = "$key1-$key2";
            next
                if exists $msg_found{$key};

            $msg_found{$key} = undef;
            # SA found first and no type specified: mark SP found
            $msg_found{"SP-$key2"} = undef
                if    $key1 eq 'SA'
                   && index("|$msg_metaf|", "|SP-$key2|") == -1
                   && index("|$msg_metaf|", "|SA-$key2|") == -1;
        }

        push @arr, $_;
    }
    return @arr;
}

{
    my %stationInfo;

    sub query_db_station {
        my ($db, $station, $station_type) = @_;

        if (!exists $stationInfo{$station_type}{$station}) {
            my ($is_fixed, $prepared, $data);

            $is_fixed = $station_type eq 'icao' || $station_type eq 'wmo';
            if ($is_fixed) {
                $prepared = $db->prepare(
                      'SELECT name, ctry, loc'
                    .  ' FROM station_fixed'
                    . ' WHERE ' . $station_type . ' = "' . $station . '"');
            } else {
                $prepared = $db->prepare(
                      'SELECT name, ctry'
                    .  ' FROM station_coords'
                   # some buoys report BBXX: lookup ship call sign in buoy table
                    . ' WHERE stationType ' . { ship  => q{IN('S','B')},
                                                mobil => q{='M'},
                                                buoy  => q{='B'},
                                                ac    => q{='A'},
                                              }->{$station_type}
                    .   ' AND station = "' . $station . '"');
            }
            $prepared->execute();
            $data = $prepared->fetchrow_hashref();
            $prepared->finish;
            $stationInfo{$station_type}{$station} = $prepared->rows == 1
                ? { ''   => $data->{name},
                    ctry => $data->{ctry},
                    ($is_fixed ? (loc => $data->{loc}) : ())
                  }
                : {};
        }
        return $stationInfo{$station_type}{$station};
    }
}

sub query_db_msgs {
    my ($db, $station_uniq, $mode, $table, $hours,
        $rep_type, $end_date, $station, $station_type) = @_;
    my ($prepared, $key, @arr, %msg_found);
    my ($from, $till);

    $station_uniq =~ s/^/'/;
    $station_uniq =~ s/$/'/;
    $station_uniq =~ s/ /','/g;
    if (defined $end_date) {
        $from = $end_date - $hours * 3600;
        $till = $end_date;
    } else {
        $from = "(SELECT MAX(time) FROM ${table}_get) - $hours * 3600";
        $till = "(SELECT MAX(time) FROM ${table}_get)";
    }
    $prepared = $db->prepare(
              'SELECT repType, station, msg, time'
            .  " FROM ${table}_get"
            . " WHERE station IN ($station_uniq)"
            .   " AND time >  $from"
            .   " AND time <= $till"
            .     ($rep_type ? " AND repType = '$rep_type'" : '')
            .     ($station_type ? " AND stationType = '$station_type'" : '')
            . ($table eq 'metar'
               ?    ' UNION ALL'
                 . ' SELECT repType, station, msg, time'
                 .   ' FROM taf_get'
                 .  " WHERE station IN ($station_uniq)"
                 .    " AND time >  $from"
                 .    " AND time <= $till"
                 .      ($rep_type ? " AND repType = '$rep_type'" : '')
               : '')
            . ' ORDER BY 4 DESC');
    $prepared->execute();
    $station =~ s/ /|/g
        if defined $station;
    while (my $data = $prepared->fetchrow_hashref()) {
        if ($table eq 'metar' || $table eq 'taf') {
            $_ = $data->{repType} . ' ' . $data->{msg};
            s/^(S[AP] )(?:METAR|SPECI) /$1/;
            s/^(F[CT] )TAF /$1/;

            if ($mode ne 'summary') {
                my ($key1, $key2);

                /^($re_report_type_metaf) /o
                    or next;
                ($key1, $key2) = ($1, $data->{station});
                $key = "$key1-$key2";
                next
                    if exists $msg_found{$key};

                $msg_found{$key} = undef;
                # SA found first and no type specified: mark SP found
                $msg_found{"SP-$key2"} = undef
                    if    $key1 eq 'SA'
                       && index("|$station|", "|SP-$key2|") == -1
                       && index("|$station|", "|SA-$key2|") == -1;
            }
            push @arr, $_;
        } elsif ($table eq 'synop_fixed' || $table eq 'synop_coords') {
            my ($id, $report_type, $was_added);
            my ($report_type_id, $report_type_matches, $report_type_any);

            $report_type = $data->{repType};
            $id = $data->{station};

            $report_type_id = "$report_type-$id";
            $report_type_matches = index("|$station|", "|$report_type_id|") > -1;
            $report_type_any     = index("|$station|", "|$id|") > -1;

            $_ = $data->{repType} . ' ' . $data->{station} . ' ' . $data->{msg};
            if ($mode eq 'summary') {
                push @arr, $_
                    if $report_type_matches || $report_type_any;
                next;
            }

            $was_added = 0;
            if ($report_type_matches && !exists $msg_found{$report_type_id}) {
                $msg_found{$report_type_id} = undef;
                $was_added = 1;
                push @arr, $_;
            }
            if ($report_type_any && !exists $msg_found{"Sx-$id"}) {
                $msg_found{"Sx-$id"} = undef;
                push @arr, $_
                    unless $was_added;
            }
        } elsif ($table eq 'buoy') {
            $_ = 'SS ' . $data->{msg};
            /^SS $re_header_buoy \d{5} /o
                or next;
            if ($mode eq 'summary' || !exists $msg_found{$data->{station}}) {
                push @arr, $_;
                $msg_found{$data->{station}} = undef;
            }
        } else {
            $_ = 'UD ' . $data->{msg};
            /^UD $re_header_amdar $re_aircraft /o
                or next;
            if ($mode eq 'summary' || !exists $msg_found{$data->{station}}) {
                push @arr, $_;
                $msg_found{$data->{station}} = undef;
            }
        }
    }
    $prepared->finish;
    return @arr;
}

sub clean_msg {
    $_ = param shift;
    return '' unless defined $_;
    $_ = uc(substr $_, 0, 2000);
    s/[^ -~]/?/g; # replace non-ASCII characters
    s/  +/ /g;
    s/ $//;
    return $_;
}

sub usage {
    my $msg = shift;

    if ($msg eq '') {
        pod2usage -exitval => 'NOEXIT', -output => \*STDERR, -noperldoc => 1;
        print {*STDERR} <<"EOF";
Options may be merged together.  -- stops processing of options.
Space is not required between options and their arguments.
EOF
    } else {
        print {*STDERR} $msg;
    }

    print {*STDERR} <<"EOF";

For more details run
        perldoc -F $0
EOF
    exit 1;
}

sub HELP_MESSAGE { usage ''; return; } # for Getopt::Std

# don't run if loaded as a library
return 1
    if caller;

{
    my %opts;

    # get and check options
    $Getopt::Std::STANDARD_HELP_VERSION = 1;
    usage '' unless getopts 'v', \%opts;
    if (exists $opts{v}) {
        print "metaf.pl version $VERSION\n" or exit 1;
        exit 0;
    }
}

=pod

If there are any CGI errors the type of error is reported and the script
terminates.

=cut

charset('UTF-8');

if (cgi_error()) {
    print header({ status => cgi_error() }) .
        start_html({ title => 'CGI Error' }) . "\n" .
        b('CGI Error: ' . cgi_error()) . "\n" .
        end_html()
      or exit 1;
    exit 0;
}

=pod

All parameters are checked to have one of the allowed values.

=cut

sub check_param {
    my ($name, $default, @allowed) = @_;
    my $val;

    $val = param $name;
    return   defined $val && { map { $_ => 1 } @allowed }->{$val} ? $val
           :                                                        $default;
}

#
# determine message types
#
my $type_metaf = check_param qw(type_metaf metar taf icao);
my $type_synop = check_param qw(type_synop synop wmo ship mobil);
my $type_buoy  = check_param qw(type_buoy  buoy wmo);
my $type_amdar = check_param qw(type_amdar amdar ac);

=pod

If the parameter C<lang> is not set the first allowed value of the HTTP header
C<Accept-Language> is used.

=cut

#
# determine language: check request parameter, then browser header. default: de
#
sub check_lang {
    my @allowed = @_; # full ('en-GB') or short ('en') names
    my ($lang, $first_short);

    # if parameter is set: check full match, then short name match
    $lang = param 'lang';
    if (defined $lang) {
        return $lang
            if { map { $_ => 1 } @allowed }->{$lang};
        $lang =
              { map { substr($_, 0, 2) => $_ } @allowed }->{substr $lang, 0, 2};
        return $lang || $allowed[0]; # TODO: or continue with browser header?
    }

    # browser header: check full match while remembering first short name match
    $first_short = '';
    for (split /,/, lc(http('Accept-Language') || '')) {
        s/ //g;
        s/;.*//;
        return $_
            if { map { $_ => 1 } @allowed }->{$_};
        $first_short =
                  { map { substr($_, 0, 2) => $_ } @allowed }->{substr $_, 0, 2}
            unless $first_short;
    }
    return $first_short || $allowed[0];
}
my $lang = check_lang qw(de en es ru);

my $is_gui = defined param 'gui';

#
# determine format: check request parameter, then browser header. default: html
#
my $format = check_param qw(format html text xml msgs);
if ($format eq 'xml') {
    my $accept = http('Accept');
    if ($accept) {
        my $accept_all = ",$accept," =~ m{[ ,]\*/\*[;,]};
        my $accept_xml = ",$accept," =~ m{[ ,](?:text|application)/(?:[a-z+-]*xml[a-z+-]*|\*)[;,]};
        my $accept_xsl = ",$accept," =~ m{[ ,](?:text|application)/(?:[a-z+-]*x[sm]l[a-z+-]*|\*)[;,]};
        $format = 'html' unless $accept_all || ($accept_xml && $accept_xsl);
    }
}

my $is_cmdline = server_software() eq 'cmdline' ? 1 : 0;
my $need_html  = !$is_cmdline || $format eq 'html';

open STDERR, '>&STDOUT'
    unless $is_cmdline;

#
# determine src_synop, src_metaf, src_buoy, src_amdar: check request parameter
#
my ($src_metaf, $src_synop, $src_buoy, $src_amdar, $had_dbi, $db);

for (split / /, $DB_DATA_SOURCES) {
    my ($forced, $data_source, $rc, $fh);

    if (!$had_dbi) {
        $rc = eval { require DBI; };
        if (!$rc || $@) {
            print {*STDERR} $@
                if $is_cmdline;
            last;
        }
        $had_dbi = 1;
    }

    $forced = s{/f$}{};

    -f "$CGI_DATA_DIR/$_.db" or ($forced ? last : next);

    if ($_ eq 'sqlite') {
        $data_source = "DBI:SQLite:$CGI_DATA_DIR/sqlite.db";
    } else {
        open $fh, '<', "$CGI_DATA_DIR/$_.db"
            or ($forced ? last : next);

        $data_source = <$fh>;
        close $fh;
        ($forced ? last : next)
            unless defined $data_source;
        chomp $data_source;
    }

    $rc = eval {
        $db = DBI->connect($data_source, 'metaf2xml', undef,
                         { AutoCommit => 1, RaiseError => 1, PrintError => 0 });
        1;
    };
    if (!$rc || $@) {
        print {*STDERR} $@
            if $is_cmdline;
        ($forced ? last : next);
    }
}

# if DB_DATA_SOURCES was forced but connecting to the DB failed: error
if ($DB_DATA_SOURCES =~ m{/f\b} && !$db) {
    cgi_error('503 Service Unavailable');
    print header({ status => cgi_error() })
        unless $is_cmdline;
    if ($need_html) {
        print start_html({ title => 'CGI Error' }) . "\n" .
            b('CGI Error: ' . cgi_error()) . "\n" .
            p('could not connect to database') .
            end_html();
    } else {
        print "Error: could not connect to database\n";
    }
    exit 0;
}

if ($db || -f "$CGI_DATA_DIR/metaf.txt") {
    $src_metaf = 'local';
} else {
    $src_metaf = check_param qw(src_metaf noaa nws adds addsds ogimet cod fsu isu);
}

if ($db || -f "$CGI_DATA_DIR/synop.txt") {
    $src_synop = 'local';
} else {
    $src_synop = check_param qw(src_synop nws ogimet cod fsu isu);
}

if ($db || -f "$CGI_DATA_DIR/buoy.txt") {
    $src_buoy = 'local';
} else {
    $src_buoy = check_param qw(src_buoy nws ogimet isu);
}

if ($db || -f "$CGI_DATA_DIR/amdar.txt") {
    $src_amdar = 'local';
} else {
    $src_amdar = check_param qw(src_amdar nws fsu isu);
}

#
# determine mode: check request parameter. default: latest
#
my $mode = check_param qw(mode latest summary);

#
# determine number of hours (for mode summary): check request param. default: 24
#
my $hours = param('hours') || '';
$hours = 24
    unless $hours =~ /^(?:[1-9]\d?|0\.\d+)$/ && $hours <= 72;

#
# determine end date for data from Ogimet, DB: check request param. default: now
#
my $end_date;
my $end_date_str = param 'end_date';
if (   defined $end_date_str
    && $end_date_str =~ /^(1999|20\d\d)-(0[1-9]|1[012])-($re_day) ($re_hour):($re_min)$/o)
{
    my @date_arr = (0, $5, $4, $3, $2, $1);
    $date_arr[5] -= 1900;                   # year is offset to 1900
    $date_arr[4]--;                         # month starts from 0
    $end_date = mktime @date_arr, -1, -1, -1;
} else {
    $end_date = undef;
    $end_date_str = '';
}

=pod

The parameters C<msg_>* are truncated to 2000 characters.
Letters are converted to upper case. Any characters that are not allowed in a
message are removed.

=cut

my $msg_metaf = clean_msg 'msg_metaf';
my $msg_synop = clean_msg 'msg_synop';
my $msg_buoy  = clean_msg 'msg_buoy';
my $msg_amdar = clean_msg 'msg_amdar';

#
# log CGI request unless called from the command line
#
if (!$is_cmdline && open my $log, '>>', $CGI_LOG_FILE) {
    # garble IP address, but still allow to group the requests
    my $addr;

    if (defined http('DNT') && substr(http('DNT'), 0, 1) eq '1') {
        $addr = 'xx.xx.xx.xx';
    } else {
        $addr = $ENV{'HTTP_X_REMOTE_ADDR'} || remote_addr();
        $addr =~ s/\.\d+/.x/;
        $addr =~ s/\d+$/x/;
    }
    print {$log} gmtime()
              . " $addr "
              . request_method()
              . " $type_metaf $type_synop $type_buoy $type_amdar $lang $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours"
              . ($is_gui ? ' gui' : '')
              . " '$msg_metaf' '$msg_synop' '$msg_buoy' '$msg_amdar'\n";
    close $log;
}

=pod

If the METAR/TAF message starts with C<METAR>, C<SPECI>, or C<TAF>, the
parameter C<type_metaf> is ignored.

=cut

#
# overwrite type_metaf if specified in message
#
$type_metaf = 'metar' if $msg_metaf =~ /^(?:METAR|SPECI) /;
$type_metaf = 'taf' if $msg_metaf =~ /^TAF /;

=pod

If the SYNOP message starts with C<AAXX>, C<BBXX> or C<OOXX>, the parameter
C<type_synop> is ignored.

=cut

#
# overwrite type_synop if specified in message
#
$type_synop = 'synop' if $msg_synop =~ /^(?:AA|BB|OO)XX /;

=pod

If the BUOY message starts with C<ZZYY>, the parameter C<type_buoy> is
ignored.

=cut

#
# overwrite type_buoy if specified in message
#
$type_buoy = 'buoy' if $msg_buoy =~ /^ZZYY /;

=pod

If the AMDAR message starts with C<AMDAR>, the parameter C<type_amdar> is
ignored.

=cut

#
# overwrite type_amdar if specified in message
#
$type_amdar = 'amdar' if $msg_amdar =~ /^$re_header_amdar /o;

#
# translated strings
#
my $trans = {
    de => {
        title => 'metaf2xml: analysiere und dekodiere METAR, TAF, SYNOP, BUOY und AMDAR',
        infosrc => 'Eine gute Quelle für aktuelle METAR- und TAF-Meldungen ist',
        infofetch => 'Wird eine Option mit "Id(s)" gewählt und eine oder mehrere Id(s) angegeben, werden Meldungen von der ausgewählten Quelle (oder, falls diese Option gesperrt ist, von dem Server, von dem diese Web-Seite kommt) geholt.',
        example => 'Beispiele für Meldungen',
        type1 => 'METAR-Meldung oder',
        type2 => 'TAF-Meldung oder',
        type3 => 'SYNOP-Meldung oder',
        type4 => 'ICAO Id(s)',
        type5 => 'WMO Id(s)',
        type6 => 'Bojen-Meldung oder',
        type7 => 'WMO Id(s) oder',
        type8 => 'SHIP Id(s) oder',
        type9 => 'MOBIL Id(s)',
        type10 => 'AMDAR-Meldung oder',
        type11 => 'Flugzeug Id(s)',
        source => 'hole Meldungen von',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS dataserver',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'Sprache',
        format => 'gib aus als',
        format1 => 'HTML',
        format2 => 'Text',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'hole/zeige',
        mode1 => 'letzte/detailliert',
        mode2 => 'alle/Übersicht seit',
        hours => 'Stunde(n)',
        decode => 'Übersetzen',
    },
    en => {
        title => 'metaf2xml: parse and decode METAR, TAF, SYNOP, BUOY and AMDAR',
        infosrc => 'A good source for up-to-date METAR and TAF messages is',
        infofetch => 'If an option with "id(s)" is chosen and one or more id(s) are given, messages will be fetched from the selected source (or if this option is disabled from the server where this web page comes from).',
        example => 'example messages',
        type1 => 'METAR message or',
        type2 => 'TAF message or',
        type3 => 'SYNOP message or',
        type4 => 'ICAO id(s)',
        type5 => 'WMO id(s)',
        type6 => 'buoy message or',
        type7 => 'WMO id(s) or',
        type8 => 'SHIP id(s) or',
        type9 => 'MOBIL id(s)',
        type10 => 'AMDAR message or',
        type11 => 'aircraft id(s)',
        source => 'fetch messages from',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS dataserver',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'language',
        format => 'output as',
        format1 => 'HTML',
        format2 => 'text',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'fetch/display',
        mode1 => 'latest/detailed',
        mode2 => 'all/summary since',
        hours => 'hour(s)',
        decode => 'decode',
    },
    es => {
        title => 'metaf2xml: analiza y decodifica METAR, TAF, SYNOP y BUOY',
        infosrc => 'Una buena fuente por mensajes actuales de METAR y TAF es',
        infofetch => 'Si es elegido una opción con "código(s)" y son dado uno o más código(s), los mensajes serán traído de la fuente elegida (o si esta opción es minusválida del servidor de donde viene esta página web).',
        example => 'ejemplos de mensajes',
        type1 => 'mensaje METAR o',
        type2 => 'mensaje TAF o',
        type3 => 'mensaje SYNOP o',
        type4 => 'código(s) OACI',
        type5 => 'código(s) OMM',
        type6 => 'mensaje boya o',
        type7 => 'código(s) OMM o',
        type8 => 'código(s) SHIP o',
        type9 => 'código(s) MOBIL',
        type10 => 'mensaje AMDAR o',
        type11 => 'código(s) aeronave',
        source => 'consultar mensajes de',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS dataserver',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'lengua',
        format => 'presentar como',
        format1 => 'HTML',
        format2 => 'texto',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'consultar/presentar',
        mode1 => 'último/detallado',
        mode2 => 'todos/resumen desde',
        hours => 'hora(s)',
        decode => 'traducir',
    },
    pl => {
        title => 'metaf2xml: METAR/TAF/SYNOP/BUOY tłumacz',
        infosrc => 'Dobrym źródłem do aktualnych raportów METAR i TAF jest',
        infofetch => 'Jeśli opcje id(s) ICAO lub id(s) WMO są wybierane i jeden lub więcej identyfikatorów stacji podano, raporty będą pobierane z wybranego źródła (lub jeśli opcja ta jest wyłączona z serwera, z którym strona ta jest).',
        example => 'przykłady raportów',
        type1 => 'raport METAR lub',
        type2 => 'raport TAF lub',
        type3 => 'raport SYNOP lub',
        type4 => 'id(s) ICAO',
        type5 => 'id(s) WMO',
        type6 => 'raport boi lub',
        type7 => 'id(s) WMO lub',
        type8 => 'id(s) SHIP lub',
        type9 => 'id(s) MOBIL',
        source => 'zapytać raporty z',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS dataserver',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'język',
        format => 'wyjście jako',
        format1 => 'HTML',
        format2 => 'tekst',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'zapytać/pokazać',
        mode1 => 'ostatni/szczegółowy',
        mode2 => 'wszystkie/podsumowanie przez',
        hours => 'godzin(y)',
        decode => 'tłumaczyć',
    },
    'pt-br' => {
        title => 'metaf2xml: METAR/TAF/SYNOP/BUOY tradutor',
        infosrc => 'Uma boa fonte de mensagens atuais METAR e TAF é',
        infofetch => 'Se a opção "código(s) OACI" ou "código(s) OMM" é escolhido e uma ou mais códigos de estação são dadas, as mensagens serão obtidos a partir da fonte selecionado (ou se esta opção for desativada a partir do servidor de onde vem esta página web).',
        example => 'exemplos de mensagens',
        type1 => 'mensagem METAR ou',
        type2 => 'mensagem TAF ou',
        type3 => 'mensagem SYNOP ou',
        type4 => 'código(s) OACI',
        type5 => 'código(s) OMM',
        type6 => 'mensagem bóia ou',
        type7 => 'código(s) OMM ou',
        type8 => 'código(s) SHIP ou',
        type9 => 'código(s) MOBIL',
        source => 'consultar mensagens de',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS dataserver',
        source4 => 'NOAA/NWS',
        source5 => 'College of DuPage',
        source6 => 'Ogimet',
        language => 'linguagem',
        format => 'presentar como',
        format1 => 'HTML',
        format2 => 'texto',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'consultar/presentar',
        mode1 => 'último/detalhado',
        mode2 => 'todos/resumo dos',
        hours => 'hora(s)',
        decode => 'traduzir',
    },
    ru => {
        title => 'metaf2xml: разбирать и декодировать METAR, TAF, SYNOP и BUOY',
        infosrc => 'Хороший источник для свежих сообщений METAR и TAF',
        infofetch => 'Если выбрана опция с "ид" и указан один или более ид, сообщения будут запрашиваться из выбранного источника (либо если эта опция отключена на сервере, с которого запрошена эта веб-страница).',
        example => 'примеры',
        type1 => 'METAR сообщение или',
        type2 => 'TAF сообщение или',
        type3 => 'SYNOP сообщение или',
        type4 => 'ICAO ид',
        type5 => 'WMO ид',
        type6 => 'буй сообщение или',
        type7 => 'WMO ид или',
        type8 => 'SHIP ид или',
        type9 => 'MOBIL ид',
        type10 => 'AMDAR сообщение или',
        type11 => 'самолет ид',
        source => 'получать сообщения от',
        source1 => 'NOAA/IWS',
        source2 => 'NOAA/ADDS',
        source3 => 'NOAA/ADDS сервер',
        source4 => 'NOAA/NWS',
        source5 => 'Колледж DuPage',
        source6 => 'Ogimet',
        language => 'язык',
        format => 'выдавать как',
        format1 => 'HTML',
        format2 => 'текст',
        format3 => 'XML (+XSL -> HTML)',
        mode => 'получить/показать',
        mode1 => 'послед./подробно',
        mode2 => 'всё/обзор с',
        hours => 'час(ов)',
        decode => 'декодировать',
    },
}->{$lang};

#
# save message parameter
#
my $msg_metaf_orig = $msg_metaf;
my $msg_synop_orig = $msg_synop;
my $msg_buoy_orig  = $msg_buoy;
my $msg_amdar_orig = $msg_amdar;

=pod

If C<type_metaf> is B<icao> and one or more airport ICAO codes are given,
messages for the specified airports are fetched from the specified source.

If C<type_synop> is B<wmo>, B<ship> or B<mobil> and one or more ids are given,
messages for the specified stations are fetched from the specified source.

If C<type_buoy> is B<wmo> and one or more WMO ids are given,
messages for the specified buoys are fetched from the specified source.

If C<type_amdar> is B<ac> and one or more aircraft ids are given,
messages for the specified aircrafts are fetched from the specified source.

=cut

#
# process METAR/TAF messages if they need to be fetched
#
if (   $type_metaf eq 'icao'
    && " $msg_metaf" =~ /^(?: (?:$re_report_type_metaf-)?$re_ICAO)+$/o)
{
    my ($icao_uniq, $icao_uniq_metar, $icao_uniq_staf, $icao_uniq_taf, @metaf);

    $icao_uniq = mk_uniq $msg_metaf =~ /(?:$re_report_type_metaf-)?([^ ]+)/og;
    $icao_uniq_metar = '';
    $icao_uniq_staf  = '';
    $icao_uniq_taf   = '';
    for ($msg_metaf =~ /((?:$re_report_type_metaf-)?[^ ]+)/og) {
        /($re_report_type_metaf-)?(.*)/o;
        if (defined $1) {
            if ($1 eq 'SA-' || $1 eq 'SP-') {
                if ($icao_uniq_metar eq '') {
                    $icao_uniq_metar = "$1$2";
                } elsif (index(" $icao_uniq_metar ", " $1$2 ") == -1) {
                    $icao_uniq_metar .= " $1$2";
                }
            } elsif ($1 eq 'FC-') {
                if ($icao_uniq_staf eq '') {
                    $icao_uniq_staf = "$1$2";
                } elsif (index(" $icao_uniq_staf ", " $1$2 ") == -1) {
                    $icao_uniq_staf .= " $1$2";
                }
            } else { # FT
                if ($icao_uniq_taf eq '') {
                    $icao_uniq_taf = "$1$2";
                } elsif (index(" $icao_uniq_taf ", " $1$2 ") == -1) {
                    $icao_uniq_taf .= " $1$2";
                }
            }
        } else {
            if ($icao_uniq_metar eq '') {
                $icao_uniq_metar = "SA-$2 SP-$2";
            } else {
                if (index(" $icao_uniq_metar ", " SA-$2 ") == -1) {
                    $icao_uniq_metar .= " SA-$2";
                }
                if (index(" $icao_uniq_metar ", " SP-$2 ") == -1) {
                    $icao_uniq_metar .= " SP-$2";
                }
            }
            if ($icao_uniq_staf eq '') {
                $icao_uniq_staf = "FC-$2";
            } elsif (index(" $icao_uniq_staf ", " FC-$2 ") == -1) {
                $icao_uniq_staf .= " FC-$2";
            }
            if ($icao_uniq_taf eq '') {
                $icao_uniq_taf = "FT-$2";
            } elsif (index(" $icao_uniq_taf ", " FT-$2 ") == -1) {
                $icao_uniq_taf .= " FT-$2";
            }
        }
    }

    if ($src_metaf eq 'noaa') {
        @metaf = (search_files_metaf($icao_uniq_metar,
                                     $mode, 'cycle_iws', $hours, 'metar'),
                  search_files_metaf($icao_uniq_staf,
                                     $mode, 'cycle_iws', $hours, 'staf'),
                  search_files_metaf($icao_uniq_taf,
                                     $mode, 'cycle_iws', $hours, 'taf'));
    } elsif ($src_metaf eq 'adds') {
        @metaf = query_metaf_adds $icao_uniq, $mode, $hours;
    } elsif ($src_metaf eq 'addsds') {
        @metaf = query_metaf_addsds($icao_uniq_metar,
                                    "$icao_uniq_staf $icao_uniq_taf",
                                    $mode, $hours, $end_date, $msg_metaf);
    } elsif ($src_metaf eq 'ogimet') {
        my $rep_type = 'ALL';
        $rep_type = $1
            if $msg_metaf =~ /^($re_report_type_metaf)-[^ ]+(?: \1-[^ ]+)*$/o;
        @metaf = query_metaf_ogimet $icao_uniq, $mode, $hours,
                                    $rep_type, $end_date, $msg_metaf;
    } elsif ({ cod => 1, fsu => 1, isu => 1 }->{$src_metaf}) {
        @metaf = search_files_metaf $icao_uniq_metar,
                                    $mode, $src_metaf, $hours, 'metaf',
                                    $end_date;
    } elsif ($src_metaf eq 'nws') {
        @metaf = (search_files_metaf($icao_uniq_metar,
                                     $mode, 'nws', $hours, 'metar'),
                  search_files_metaf("$icao_uniq_staf $icao_uniq_taf",
                                     $mode, 'nws', $hours, 'taf'));
    } elsif ($db) {
        my $rep_type = '';
        $rep_type = $1
            if $msg_metaf =~ /^($re_report_type_metaf)-[^ ]+(?: \1-[^ ]+)*$/o;
        @metaf = query_db_msgs($db, $icao_uniq, $mode, 'metar', $hours,
                               $rep_type, $end_date, $msg_metaf);
    } else { # local
        @metaf = search_files_metaf
                              "$icao_uniq_metar $icao_uniq_staf $icao_uniq_taf",
                                    $mode, $src_metaf, $hours, 'metaf';
    }

    $msg_metaf = '';
    for (@metaf) {
        if (/ERROR/) {
            $msg_metaf = $_;
            last;
        }
    }

    if (!$msg_metaf) {
        for my $id (split / /, $icao_uniq) {
            my ($ii, $type_pattern);

            $type_pattern = '';
            $ii = 0;
            $ii++    if index(" $icao_uniq_metar ", " SA-$id ") > -1;
            $ii += 2 if index(" $icao_uniq_metar ", " SP-$id ") > -1;
            $type_pattern = qw(SA SP S[AP])[$ii - 1]
                if $ii > 0;
            $ii = 0;
            $ii++    if index(" $icao_uniq_staf ", " FC-$id ") > -1;
            $ii += 2 if index(" $icao_uniq_taf ",  " FT-$id ") > -1;
            if ($ii > 0) {
                $type_pattern .= '|' if $type_pattern;
                $type_pattern .= qw(FC FT F[CT])[$ii - 1];
            }
            next unless $type_pattern;

            $msg_metaf .= join "\n", grep { /^(?:$type_pattern) $re_header_metaf$id / } @metaf;
            $msg_metaf .= "\n";
        }
        if ($msg_metaf) {
            $msg_metaf =~ s/\n+/\n/g;
            $msg_metaf =~ s/^SA (?=METAR )//mg;
            $msg_metaf =~ s/^SA /METAR\n/mg;
            $msg_metaf =~ s/^SP (?=SPECI )//mg;
            $msg_metaf =~ s/^SP /SPECI /mg;
            $msg_metaf =~ s/^F[CT] (?=TAF )//mg;
            $msg_metaf =~ s/^F[CT] /TAF\n/mg;
        }
    }
} elsif ($msg_metaf ne '') {
    $msg_metaf = ($type_metaf eq 'taf' ? 'TAF' : 'METAR') . "\n$msg_metaf";
}

# process SYNOP messages if they need to be fetched
if (   $type_synop ne 'synop'
    && (   " $msg_synop" =~ /^(?: (?:$re_report_type_synop-)?[A-Z\d]{3,})+$/o
        || $msg_synop =~ /^(?:$re_report_type_synop-)?[A-Z]+$/o))
{
    my ($wmoid_uniq, @synop);

    $wmoid_uniq = mk_uniq $msg_synop =~ /(?:$re_report_type_synop-)?([^ ]+)/og;
    if ($src_synop eq 'ogimet') {
        my $rep_type = 'ALL';
        $rep_type = $1
              if $msg_synop =~ /^($re_report_type_synop)-[^ ]+(?: \1-[^ ]+)*$/o;
        @synop = query_synop_ogimet $msg_synop, $mode, $type_synop, $wmoid_uniq,
                                    $hours, $rep_type, $end_date;
    } elsif ($db) {
        my $rep_type = '';
        $rep_type = $1
              if $msg_synop =~ /^($re_report_type_synop)-[^ ]+(?: \1-[^ ]+)*$/o;
        if ($type_synop eq 'wmo') {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_fixed',
                                   $hours, $rep_type, $end_date, $msg_synop;
        } elsif ($type_synop eq 'ship') {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_coords',
                                   $hours, $rep_type, $end_date, $msg_synop,'S';
        } else {
            @synop = query_db_msgs $db, $wmoid_uniq, $mode, 'synop_coords',
                                   $hours, $rep_type, $end_date, $msg_synop,'M';
        }
    } else { # cod, fsu, isu, nws, local
        @synop = search_files_synop $msg_synop, $mode, $src_synop, $hours,
                                    $type_synop, $end_date;
    }

    $msg_synop = '';
    for (@synop) {
        if (/ERROR/) {
            $msg_synop = $_;
            last;
        }
    }

    if (!$msg_synop) {
        if ($wmoid_uniq =~ /[A-Z]$/) {
            $msg_synop = join("\n", @synop) . "\n";
        } else {
            for my $id (split / /, $wmoid_uniq) {
                $msg_synop .=
                       join("\n", grep { /^$re_report_type_synop $id / } @synop)
                     . "\n";
            }
        }
        if ($msg_synop) {
            $msg_synop =~ s/^$re_report_type_synop [A-Z\d]+ //omg;
            $msg_synop = "\n$msg_synop";
        }
    }
}

# process BUOY messages if they need to be fetched
if ($type_buoy eq 'wmo' && " $msg_buoy" =~ /^(?: \d{5})+$/) {
    my ($wmoid_uniq, @buoy);

    $wmoid_uniq = mk_uniq $msg_buoy =~ /([^ ]+)/g;
    if ($src_buoy eq 'ogimet') {
        @buoy = query_buoy_ogimet $mode, $wmoid_uniq, $hours, 'ALL', $end_date;
    } elsif ($db) {
        @buoy = query_db_msgs $db, $wmoid_uniq, $mode, 'buoy', $hours, '',
                              $end_date;
    } else { # nws, isu, local
        @buoy = search_files_buoy $msg_buoy, $mode, $src_buoy, $hours,
                                  $end_date;
    }

    $msg_buoy = '';
    for (@buoy) {
        if (/ERROR/) {
            $msg_buoy = $_;
            last;
        }
    }

    if (!$msg_buoy) {
        for my $id (split / /, $wmoid_uniq) {
            $msg_buoy .=
                        join("\n", grep { /^SS $re_header_buoy $id / } @buoy)
                      . "\n";
        }
        if ($msg_buoy) {
            $msg_buoy =~ s/^SS //mg;
            $msg_buoy = "\n$msg_buoy";
        }
    }
}

# process AMDAR messages if they need to be fetched
if ($type_amdar eq 'ac' && " $msg_amdar" =~ /^(?: $re_aircraft)+$/) {
    my ($acid_uniq, @amdar);

    $acid_uniq = mk_uniq $msg_amdar =~ /([^ ]+)/g;
    if ($db) {
        @amdar = query_db_msgs $db, $acid_uniq, $mode, 'amdar', $hours, '',
                               $end_date;
    } else { # nws, fsu, isu, local
        @amdar = search_files_amdar $msg_amdar, $mode, $src_amdar, $hours,
                                    $end_date;
    }

    $msg_amdar = '';
    for (@amdar) {
        if (/ERROR/) {
            $msg_amdar = $_;
            last;
        }
    }

    if (!$msg_amdar) {
        for my $id (split / /, $acid_uniq) {
            $msg_amdar .=
                         join("\n", grep { /^UD $re_header_amdar $id / } @amdar)
                         . "\n";
        }
        if ($msg_amdar) {
            $msg_amdar =~ s/^UD //mg;
            $msg_amdar = "\n$msg_amdar";
        }
    }
}

my $msg = $msg_metaf . "\n" . $msg_synop . "\n" . $msg_buoy . "\n" . $msg_amdar;
$msg =~ s/[^\n -~]/?/g; # replace non-ASCII characters
$msg =~ s/  +/ /g;
$msg =~ s/ $//mg;
$msg =~ s/^\n//mg;

if ($format eq 'msgs') {
    print header('text/plain')
        unless $is_cmdline;
    print $msg;
    exit 0;
}

$metaf2xml::parser::_mk_station_info = sub { query_db_station $db, @_; }
    if $db && !$is_gui;

if ($format eq 'xml') {
    my $msg_type;

    my $xsl = $mode eq 'summary' ? 'sum-ui' : 'fullhtml';
    start_xml { o => '-',
                $is_cmdline ? () : (I => header('application/xml')),
                S => "../metaf-$xsl.xsl",
                O => "$type_metaf $type_synop $type_buoy $type_amdar $lang $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours $msg_metaf_orig  $msg_synop_orig  $msg_buoy_orig  $msg_amdar_orig"
    };
    $msg_type = 'METAR';
    for (split /\n/, $msg) {
        if ({ METAR => 1, SPECI => 1, TAF => 1 }->{$_}) {
            $msg_type = $_;
            next;
        }
        parse_report $_, $msg_type;
    }
    finish;
    exit 0;
}

print header()
    unless $is_cmdline;
print start_html({ title => $trans->{title},
                   lang  => $lang,
                   dtd   => [ '-//W3C//DTD HTML 4.01 Transitional//EN',
                              'http://www.w3.org/TR/html4/loose.dtd' ],
                   meta  => { robots   => 'noindex,nofollow',
                              keywords =>
          'metaf2xml,parse,decode,METAR,SPECI,SAO,TAF,SYNOP,BUOY,AMDAR,XML,aviation,weather,report,forecast',
                              viewport => 'width=device-width, initial-scale=1',
                          },
                   head  => [ meta({ http_equiv => 'Content-Type',
                                     content    => 'text/html; charset=UTF-8'}),
                              Link({ rel        => 'shortcut icon',
                                     type       => 'image/x-icon',
                                     href       => '../metaf2xml.ico'})
                            ],
                   style => { src => '../metaf2xml.css' },
               }
              ) . "\n" .
    h1($trans->{title}) . "\n"
        if $need_html;

print
    p($trans->{infosrc} . ":\n" .
      kbd('http://aviationweather.gov/dataserver')) . "\n" .
    p($trans->{infofetch}) . "\n" .
    start_form({ action => 'metaf.pl', method => 'GET' }) .
    table({ border => 0, cellspacing => 0 }, "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans->{language} . ': ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'de',
                      ($lang eq 'de' ? 'checked' : '') => undef }
                   ) .
              'Deutsch ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'en',
                      ($lang eq 'en' ? 'checked' : '') => undef }
                   ) .
              'English ') . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'es',
                      ($lang eq 'es' ? 'checked' : '') => undef }
                   ) .
              'Español ') . "\n" .
           td(comment(input({ type => 'radio', name => 'lang', value => 'pl',
                      ($lang eq 'pl' ? 'checked' : '') => undef }
                   ) .
              'Polski ')) . "\n" .
           td(comment(input({ type => 'radio', name => 'lang', value => 'pt-br',
                      ($lang eq 'pt-br' ? 'checked' : '') => undef }
                   ) .
              'Português brasileiro ')) . "\n" .
           td(input({ type => 'radio', name => 'lang', value => 'ru',
                      ($lang eq 'ru' ? 'checked' : '') => undef }
                   ) .
              'Русский ') . "\n"
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans->{format} . ': ') . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'html',
                      ($format eq 'html' ? 'checked' : '') => undef }
                   ) .
              $trans->{format1} . ' ') . "\n" .
           td(input({ type => 'radio', name => 'format', value => 'text',
                      ($format eq 'text' ? 'checked' : '') => undef }
                   ) .
              $trans->{format2} . ' ') . "\n" .
           td($ENV{M2X_TEST} ? () : { colspan => 4 },
              input({ type => 'radio', name => 'format', value => 'xml',
                      ($format eq 'xml' ? 'checked' : '') => undef }
                   ) .
              $trans->{format3} . ' ') . "\n" .
           ($ENV{M2X_TEST} ?
           td({ colspan => 3 },
              input({ type => 'radio', name => 'format', value => 'msgs',
                      ($format eq 'msgs' ? 'checked' : '') => undef }
                   ) .
              'raw ') . "\n"
           : '')
        ) . "\n" .
        Tr("\n" .
           td({ nowrap => undef }, $trans->{mode} . ': ') . "\n" .
           td({ nowrap => undef },
              input({ type => 'radio', name => 'mode', value => 'latest',
                      ($mode eq 'latest' ? 'checked' : '') => undef }
                   ) .
              $trans->{mode1} . ' ') . "\n" .
           td({ $ENV{M2X_TEST} ? () : (colspan => 5), nowrap => undef },
              input({ type => 'radio', name => 'mode', value => 'summary',
                      ($mode eq 'summary' ? 'checked' : '') => undef }
                   ) .
              $trans->{mode2} . "\n" .
              textfield({ name      => 'hours',
                          size      => 3,
                          maxlength => 3,
                          class     => 'mono',
                          default   => $hours,
                          override  => 1}) . ' ' .
              $trans->{hours} . ' '
             ) . "\n" .
           ($ENV{M2X_TEST} ?
           td({ colspan => 4, nowrap => undef },
              'before ',
              textfield({ name      => 'end_date',
                          size      => 16,
                          maxlength => 16,
                          class     => 'mono',
                          default   => $end_date_str,
                          override  => 1}) .
              ' (YYYY-MM-DD hh:mm)'
              ) . "\n"
           : '')
        ) . "\n"
    ) . "\n" .
    table({ border => 0, cellspacing => 0, width => '100%' }, "\n" .
       Tr(td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top', rowspan => 3 }, "\n" .
             $trans->{example} . ': ') . "\n" .
          td({ class => 'mono' }, "\n" .
             'METAR YUDO 090600Z 00000KT CAVOK 22/15 Q1021 NOSIG') . "\n"
       ) . "\n" .
       Tr("\n" .
          td({ class => 'mono' }, "\n" .
             'TAF YUDO 090600Z 0907/0916 VRB03KT 8000 SKC PROB40 TEMPO 0907/0908 0200 +TSRA FM090800 CAVOK') . "\n"
       ) . "\n" .
       Tr("\n" .
          td({ class => 'mono' }, "\n" .
             'TAF YUDO 090600Z 090716 VRB03KT 8000 SKC PROB40 TEMPO 0708 0200 +TSRA FM0800 CAVOK') . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_metaf', value => 'metar',
                   ($type_metaf eq 'metar' ? 'checked' : '') => undef }
                ) .
           $trans->{type1} . " \n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_metaf', value => 'taf',
                   ($type_metaf eq 'taf' ? 'checked' : '') => undef }
                ) .
           $trans->{type2} . " \n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_metaf', value => 'icao',
                   ($type_metaf eq 'icao' ? 'checked' : '') => undef }
                ) .
           $trans->{type4} . ": \n"
        ). "\n" .
        td({ valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_metaf',
                       size      => 80,
                       maxlength => 2000,
                       class     => 'mono',
                       default   => $msg_metaf_orig,
                       override  => 1}) . "\n") . "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans->{source} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_metaf', value => 'noaa',
                     ($src_metaf eq 'noaa' ? 'checked' : '') => undef,
                     ($src_metaf eq 'local' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source1} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'nws',
                     ($src_metaf eq 'nws' ? 'checked' : '') => undef,
                     ($src_metaf eq 'local' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source4} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'adds',
                     ($src_metaf eq 'adds' ? 'checked' : '') => undef,
                     ($src_metaf eq 'local' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source2} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'addsds',
                     ($src_metaf eq 'addsds' ? 'checked' : '') => undef,
                     ($src_metaf eq 'local' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source3} . "\n" .
             input({ type => 'radio', name => 'src_metaf', value => 'ogimet',
                     ($src_metaf eq 'ogimet' ? 'checked' : '') => undef,
                     ($src_metaf eq 'local' ? 'readonly' : '') => undef,
                     ($src_metaf eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source6} . ' '
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top', rowspan => 3 }, "\n" .
             $trans->{example} . ': ') . "\n" .
          td({ class => 'mono' }, "\n" .
             'AAXX 09004 08495 11459 30714 10147 20136 30151 40159 58005 60001 70511 83500 92350') . "\n"
       ) . "\n" .
       Tr("\n" .
          td({ class => 'mono' }, "\n" .
             'BBXX UIDS 18061 99613 70284 41698 61806 10000 40107 57020 70121 86//8 22271 00064') . "\n"
       ) . "\n" .
       Tr("\n" .
          td({ class => 'mono' }, "\n" .
             'OOXX AAERC 10104 99815 31639 60713 00451 46/// /1511 11185 39739 49798 5//// 90940') . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_synop', value => 'synop',
                   ($type_synop eq 'synop' ? 'checked' : '') => undef }
                ) .
           $trans->{type3} . " \n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_synop', value => 'wmo',
                      ($type_synop eq 'wmo' ? 'checked' : '') => undef }
                   ) .
           $trans->{type7} . "\n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_synop', value => 'ship',
                      ($type_synop eq 'ship' ? 'checked' : '') => undef }
                   ) .
           $trans->{type8} . "\n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_synop', value => 'mobil',
                      ($type_synop eq 'mobil' ? 'checked' : '') => undef }
                   ) .
           $trans->{type9} . ": \n"
        ). "\n" .
        td({ valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_synop',
                       size      => 80,
                       maxlength => 2000,
                       class     => 'mono',
                       default   => $msg_synop_orig,
                       override  => 1}) . "\n") . "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans->{source} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_synop', value => 'nws',
                     ($src_synop eq 'nws' ? 'checked' : '') => undef,
                     ($src_synop eq 'local' ? 'readonly' : '') => undef,
                     ($src_synop eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source4} . "\n" .
             input({ type => 'radio', name => 'src_synop', value => 'cod',
                     ($src_synop eq 'cod' ? 'checked' : '') => undef,
                     ($src_synop eq 'local' ? 'readonly' : '') => undef,
                     ($src_synop eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source5} . "\n" .
             input({ type => 'radio', name => 'src_synop', value => 'ogimet',
                     ($src_synop eq 'ogimet' ? 'checked' : '') => undef,
                     ($src_synop eq 'local' ? 'readonly' : '') => undef,
                     ($src_synop eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source6} . ' '
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top', rowspan => 1 }, "\n" .
             $trans->{example} . ': ') . "\n" .
          td({ class => 'mono' }, "\n" .
             'ZZYY 47503 02031 2206/ 769820 097173 6112/ 444 20120 02031 2201/ 500// 9/000') . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_buoy', value => 'buoy',
                   ($type_buoy eq 'buoy' ? 'checked' : '') => undef }
                ) .
           $trans->{type6} . " \n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_buoy', value => 'wmo',
                   ($type_buoy eq 'wmo' ? 'checked' : '') => undef }
                ) .
           $trans->{type5} . ": \n"
        ). "\n" .
        td({ valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_buoy',
                       size      => 80,
                       maxlength => 2000,
                       class     => 'mono',
                       default   => $msg_buoy_orig,
                       override  => 1}) . "\n") . "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans->{source} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_buoy', value => 'nws',
                     ($src_buoy eq 'nws' ? 'checked' : '') => undef,
                     ($src_buoy eq 'local' ? 'readonly' : '') => undef,
                     ($src_buoy eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source4} . "\n" .
             input({ type => 'radio', name => 'src_buoy', value => 'ogimet',
                     ($src_buoy eq 'ogimet' ? 'checked' : '') => undef,
                     ($src_buoy eq 'local' ? 'readonly' : '') => undef,
                     ($src_buoy eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source6} . ' '
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ valign => 'top', rowspan => 1 }, "\n" .
             $trans->{example} . ': ') . "\n" .
          td({ class => 'mono' }, "\n" .
             'AMDAR 1605 ASC AFZA51 2624S 02759E 160414 F156 MS048 278/034 TB0 S031 333 F/// VG010') . "\n"
       ) . "\n" .
       Tr("\n" .
        td({ nowrap => undef }, "\n" .
           input({ type => 'radio', name => 'type_amdar', value => 'amdar',
                   ($type_amdar eq 'amdar' ? 'checked' : '') => undef }
                ) .
           $trans->{type10} . " \n" .
           br() . "\n" .
           input({ type => 'radio', name => 'type_amdar', value => 'ac',
                   ($type_amdar eq 'ac' ? 'checked' : '') => undef }
                ) .
           $trans->{type11} . ": \n"
        ). "\n" .
        td({ valign => 'bottom' }, "\n" .
           textfield({ name      => 'msg_amdar',
                       size      => 80,
                       maxlength => 2000,
                       class     => 'mono',
                       default   => $msg_amdar_orig,
                       override  => 1}) . "\n") . "\n"
       ). "\n" .
       Tr("\n" .
          td({ nowrap => undef, align => 'right' },
             $trans->{source} . ': ') . "\n" .
          td(input({ type => 'radio', name => 'src_amdar', value => 'nws',
                     ($src_amdar eq 'nws' ? 'checked' : '') => undef,
                     ($src_amdar eq 'local' ? 'readonly' : '') => undef,
                     ($src_amdar eq 'local' ? 'disabled' : '') => undef,
                   }
                  ) .
             $trans->{source4} . "\n"
          ) . "\n"
       ) . "\n" .
       Tr("\n" . td({ colspan => 2 }, hr)) . "\n" .
       Tr("\n" .
          td({ colspan => 2 },
            submit({ class => 'button', value => $trans->{decode} }))
          . "\n")
    ) . "\n" .
    end_form . "\n"
        unless $is_cmdline;

if ($msg_metaf or $msg_synop or $msg_buoy or $msg_amdar) {
    my ($msg_type, $sum, $fh);

    $sum = $mode eq 'summary' ? 'sum-' : '';
    print p() . '<div class="output">' . start_table() . '<tr><td><pre>'
        if !$is_cmdline && $format eq 'text';
    if (!pipe_open \$fh, 0, qw(xsltproc --nonet --nowrite --nomkdir --stringparam lang), $lang, qw(--stringparam no_links), $is_cmdline, "$METAF2XML_XSL/metaf-$sum$format.xsl", '-')
    {
        my $error_info = 'ERROR starting xsltproc';
        print $need_html ? p(b($error_info)) . "\n" : "$error_info\n";
    } else {
        start_xml { o => '&' . fileno $fh,
                    O => "$type_metaf $type_synop $type_buoy $type_amdar $lang $format $src_metaf $src_synop $src_buoy $src_amdar $mode $hours $msg_metaf_orig  $msg_synop_orig  $msg_buoy_orig  $msg_amdar_orig"
        };
        $msg_type = 'METAR';
        for (split /\n/, $msg) {
            if ({ METAR => 1, SPECI => 1, TAF => 1 }->{$_}) {
                $msg_type = $_;
                next;
            }
            parse_report $_, $msg_type;
        }
        finish;
        if (!close $fh || $?) {
            my $exit_info = exit_info($!, $?) . 'xsltproc';
            print $need_html ? p(b($exit_info)) . "\n" : "$exit_info\n";
        }
    }
    print '</pre></td></tr>' . end_table() . '</div>'
        if !$is_cmdline && $format eq 'text';
} elsif ($need_html) {
    print comment("Perl: $^V"), "\n", comment(scalar gmtime), "\n";
}

print $need_html
        ? comment('metaf.pl: 1.58') . "\n" .
          p("copyright (c) 2006-2015 metaf2xml @\n" .
            a({ href => 'http://metaf2xml.sourceforge.net/' }, '
            SourceForge.net
          ')) .
          end_html()
        : 'metaf.pl: 1.58' .
           " copyright (c) 2006-2015 metaf2xml @ SourceForge.net\n";

exit 0;

=head1 EXIT STATUS

If an invalid option was provided, the script will exit with status 1, otherwise
it will exit with status 0.

=head1 EXAMPLES

Parse a METAR message and print the result as HTML:

 metaf.pl msg_metaf="YUDO 090600Z 00000KT CAVOK 22/15 Q1021"

Fetch up-to-date METAR and TAF messages (from NOAA/IWS), SYNOP messages (from
NOAA/NWS land) for Rio (SBGL, 83746), New York (KJFK, 74486), and Tokyo
(RJTT, 47662) and BUOY messages from a buoy nearby (from NOAA/NWS) and print
the result as XML:

 metaf.pl format=xml type_metaf=icao type_synop=wmo type_buoy=wmo msg_metaf="SBGL KJFK RJTT" msg_synop="83746 74486 47662" msg_buoy="31550 41997 53596"

=head1 SEE ALSO

L<metaf2xml::parser|metaf2xml::parser>(3pm),
L<metaf2xml::src2raw|metaf2xml::src2raw>(3pm),
L<metafsrc2raw|metafsrc2raw>(1),
L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2006-2015 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
L<http://www.gnu.org/licenses/>.

=cut
