########################################################################
# metaf2xml/bufr.pm 1.58
#   process a decoded BUFR message
#
# copyright (c) 2012-2015 metaf2xml
#
# This file is part of metaf2xml.
#
# metaf2xml is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# metaf2xml is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with metaf2xml; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
# <http://www.gnu.org/licenses/>.
########################################################################

package metaf2xml::bufr;

########################################################################
# some things strictly Perl
########################################################################
use strict;
use warnings;
# TODO: use -$x instead of 0-$x with Perl 5.14.0+
use 5.010_001;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    require Exporter;

    our @ISA       = qw(Exporter);
    our @EXPORT_OK = qw();
}

END {}

our $VERSION = 1.058;
sub VERSION {
    my ($pkg, $vers) = @_;

    die "$pkg version $vers required--this is version $VERSION\n"
        if $#_ == 1 && $vers != $VERSION;
    return $VERSION;
}

=head1 NAME

metaf2xml::bufr - process a decoded BUFR message

=head1 SYNOPSIS

 use metaf2xml::bufr 1.058;

 metaf2xml::bufr::_bufr2report(\@data,\@desc);

=head1 DESCRIPTION

This Perl module provides a function to process a decoded BUFR message.

=cut

sub _mk_s {
    my ($data, $desc, $start, $cnt) = @_;

    return join ' ', map { $desc->[$_] . ':' . ($data->[$_] // '/') }
                         ($start .. $start + $cnt - 1);
}

sub _K2C {
    return { v => sprintf('%.1f', (shift) - 273.15), u => 'C' };
}

sub _addRadiationSun {
    my ($report, $rad) = @_;

    for (@{$report->{section3}}) {
        if (exists $_->{radiationSun}) {
            if (   $_->{radiationSun}{sunshinePeriod}{v}
                == $rad->{sunshinePeriod}{v})
            {
                $_->{radiationSun}{s} .= ' ' . $rad->{s};
                delete $rad->{s};
                $_ = { radiationSun => { %{$_->{radiationSun}}, %$rad }};
                delete $_->{radiationSun}{sunshineNotAvailable}
                    if exists $_->{radiationSun}{sunshine};
                return;
            }
        }
    }
    push @{$report->{section3}}, { radiationSun => $rad };
    return;
}

=head1 SUBROUTINES/METHODS

=cut

########################################################################
# _bufr2report
#
# processed groups (categories 0 and 1):
# 001003 001020 001005 002001 002036 002149 301011 301012 008021 301011 301012 008021 301021 027004 028004 007030 001051 002148 001012 001014 002040 033022 033023 033027 022063 302021 302022 302023 008081 025026 008081 025026 008081 025026 008081 002034 022060 007070 002190 025086 002035 002168 020031 002038 306004 002030 306005 007031 008081 012064 302001 008081 007032 007033 012101 012103 013003 007032 007033 008082 007033 002169 002002 008021 004025 011001 011002 008021 004025 011043 011041 008082 007033 007032 004024 013011 007032 008021 004024 014021 008021 025028 025028 025028
#  category: 1 (surface data - sea)
#  int. subcategory: 25 (Buoy observation (BUOY))
#  obs time:
#  301011: 004001 004002 004003
#  301012: 301012 301012
#  last known position:
#  008021=26 301011 301012 008021=/
#  301021: 005001 006001
#  ...
# 301001 001015 002001 301021 007030 007031 118000 031001 301011 301012 004025 010004 007032 012101 012103 013003 007032 004025 013011 007032 002002 008021 004025 011001 011002 008021
# 307022
# 307079 004025 011042
# 307080
#  category: 0 (Surface data - land)
#  int. subcategory: 2 (Main synoptic observations from fixed-land stations (SYNOP))
# 307086
#  category: 0 (Surface data - land)
#  int. subcategory: 2 (Main synoptic observations from fixed-land stations (SYNOP))
# 307086 001023 004025 002177 101000 031001 020003 103000 031001 005021 020001 005021 101000 031000 302056 103000 031000 033041 020058 022061 101000 031000 302022 101000 031001 302023 103000 031001 020054 020012 020090 004025 013012 004025 011042 104000 031001 008021 004025 011042 008021 115000 031001 008021 004015 008021 004025 011001 011002 008021 004015 008021 004025 011001 011002 008021 004025 004015 103000 031001 004025 004025 020003 111000 031001 004025 004025 005021 005021 020054 020024 020025 020026 020027 020063 008021
# 307091
# 307096
# 308004
# 308009
########################################################################
sub _bufr2report {

=head2 _bufr2report(\@data,\@desc)

This function processes a decoded BUFR message. (PARTIALLY IMPLEMENTED.)
It is a helper function used by C<metaf2xml::parser>.

The following arguments are expexted:

=over

=item data

an array of values for a message subset

=item desc

an array of descriptors with the full hierarchy of groups for the same message
subset

=back

=cut

    my ($data, $desc) = @_;
    my (%report, $not_processed_s, $ii, $is_processed, $val0, $s0, $is_auto);
    local $_;

    %report = (msg => _mk_s $data, $desc, 0, $#$desc + 1);

    $_ = "@$desc";
    $ii = 0;
    $not_processed_s = '';
    while ($ii <= $#$desc) {
        $is_processed = 1;
        pos   = 7 * $ii;
        $val0 = $data->[$ii];
        $s0   = _mk_s $data, $desc, $ii, 1;

        if (   /\G001011 001012 001013/
                 && defined $val0
                 && !exists $report{callSign})
        {
            $report{callSign} = { s => $s0, id => $val0 };
            $report{isSynop} = undef;
            @{$report{obsStationType}}{qw(s stationType)} = qw(BBXX BBXX);
            if (defined $data->[$ii + 1] && defined $data->[$ii + 2]) {
                $report{displacement} = {
                    s             => _mk_s($data, $desc, $ii, 3),
                    timeBeforeObs => { notAvailable => undef },
                      $data->[$ii + 2] == 0   ? (isStationary => undef)
                    : $data->[$ii + 1] == 509 ? (dirVarAllUnk => undef)
                    :                           (dir => $data->[$ii + 1])
                };
                if (!exists $report{displacement}{isStationary}) {
                    $report{displacement}{speed} = [
                        { v => $data->[$ii + 2], u => 'MPS' },
                        { v => $data->[$ii + 2], u => 'MPS' }
                    ];
                }
            }
            $ii += 3;
        } elsif (   /\G001003 001020 001005/
                 && defined $val0
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2]
                 && !exists $report{buoyId})
        {
            $report{buoyId} = {
                s  => _mk_s($data, $desc, $ii, 3),
                id => sprintf '%d%d%03d', @$data[$ii .. $ii + 2]
            };
            $report{isBuoy} = undef;
            @{$report{obsStationType}}{qw(s stationType)} = qw(ZZYY ZZYY);
            $ii += 3;
        } elsif (  /\G004001 004002 004003 004004 004005/
                 && defined $val0
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2]
                 && defined $data->[$ii + 3]
                 && defined $data->[$ii + 4])
        {
            if (exists $report{obsTime}) {
                # another observation!
                $not_processed_s .= ' '
                                  . _mk_s($data, $desc, $ii, $#$desc - $ii + 1);
                $ii = $#$desc + 1;
            } else {
                $report{obsTime} = {
                    s  => _mk_s($data, $desc, $ii, 5),
                    timeAt => {
                        year   => $val0,
                        month  => $data->[$ii + 1],
                        day    => $data->[$ii + 2],
                        hour   => sprintf('%02d', $data->[$ii + 3]),
                        minute => sprintf('%02d', $data->[$ii + 4])
                    }
                };
                $ii += 5;
            }
        } elsif (/\G002001/) {
            if (defined $val0 && $val0 == 0) {
                $report{reportModifier} = { s => $s0, modifierType => 'AUTO' };
                $is_auto = 'AUTO';
            }
            $ii++;
        } elsif (   (/\G005001 006001/ || /\G005002 006002/)
                 && !(   exists $report{stationPosition}
                      && exists $report{stationPosition}{lat}))
        {
            $report{stationPosition}{s} = _mk_s($data, $desc, $ii, 2);
            if (defined $val0 && defined $data->[$ii + 1]) {
                delete $report{stationPosition}{notAvailable};
                @{$report{stationPosition}}{qw(lat lon)} = (
                                      { v => $data->[$ii],     q => 10000 },
                                      { v => $data->[$ii + 1], q => 10000 }
                );
            } else {
                $report{stationPosition}{notAvailable} = undef;
            }
            $ii += 2;
        } elsif (/\G007030/
                 && !(   exists $report{stationPosition}
                      && exists $report{stationPosition}{elevation}))
        {
            if (exists $report{stationPosition}) {
                $report{stationPosition}{s} .= " $s0";
                $report{stationPosition}{elevation} = { v => $val0, u => 'M' }
                    if    defined $val0
                       && exists $report{stationPosition}{lat};
            } else {
                $report{stationPosition} = {
                    s            => $s0,
                    notAvailable => undef
                };
            }
            $ii++;
        } elsif (   /\G001001 001002/
                 && defined $val0
                 && defined $data->[$ii + 1]
                 && !exists $report{obsStationId})
        {
            $report{obsStationId} = {
                s  => _mk_s($data, $desc, $ii, 2),
                id => sprintf '%02d%03d', @$data[$ii .. $ii + 1]
            };
            $report{isSynop} = undef;
            @{$report{obsStationType}}{qw(s stationType)} = qw(AAXX AAXX);
            $ii += 2;

        } elsif (   /\G008021 004001 004002 004003 004004 004005 008021/
                 && defined $val0
                 && defined $data->[$ii + 1]
                 && defined $data->[$ii + 2]
                 && defined $data->[$ii + 3]
                 && defined $data->[$ii + 4]
                 && defined $data->[$ii + 5]
                 && !defined $data->[$ii + 6]
                 && $val0 == 26
                 && !exists $report{lastKnownPosTime})
        {
            $report{lastKnownPosTime} = {
                s  => _mk_s($data, $desc, $ii, 7),
                timeAt => {
                    year   => $data->[$ii + 1],
                    month  => $data->[$ii + 2],
                    day    => $data->[$ii + 3],
                    hour   => sprintf('%02d', $data->[$ii + 4]),
                    minute => sprintf('%02d', $data->[$ii + 5])
                }
            };
            $ii += 7;
        } elsif (  /\G007062 022043 022062/
                 && defined $val0
                 && $val0 == 0
                 && defined $data->[$ii + 1]
                 && !defined $data->[$ii + 2]
                 && !exists $report{seaSurfaceTemp})
        {
            $report{seaSurfaceTemp} = {
                s    => _mk_s($data, $desc, $ii, 3),
                temp => _K2C $data->[$ii + 1]
            };
            $ii += 3;
        } elsif (   exists $report{isSynop}
                 && /\G007004 010009/
                 && !exists $report{gpSurface}
                 && defined $val0 && defined $data->[$ii + 1])
        {
            if (exists $report{SLP}) {
                $report{ERROR} = { errorType => 'other', s => 'SLP' };
                last;
            }
            if ({ map { $_ => 1 } qw(1000 925 500 700 850) }->{$val0 / 100}) {
                $report{gpSurface} = {
                    s            => _mk_s($data, $desc, $ii, 2),
                    surface      => $val0 / 100,
                    geopotential => $data->[$ii + 1]
                };
            } else {
                $report{gpSurface} = {
                    s             => _mk_s($data, $desc, $ii, 2),
                    invalidFormat => _mk_s($data, $desc, $ii, 2)
                };
            }
            $ii += 2;
        } elsif (/\G012101 012103 013003/ && !exists $report{temperature}) {
            $report{temperature} = {
                s         => _mk_s($data, $desc, $ii, 3),
                air       => defined $val0 ? { temp => _K2C $val0 }
                                           : { notAvailable => undef },
                dewpoint  => defined $data->[$ii + 1]
                                    ? { temp => _K2C $data->[$ii + 1] }
                                    : { notAvailable => undef },
                defined $data->[$ii + 2]
                        ? (relHumid1 => $data->[$ii + 2])
                        : ()
            };
            $ii += 3;
        } elsif (/\G010004/ && defined $val0) {
            $report{stationPressure} = {
                s        => $s0,
                pressure => { v => sprintf('%.1f', $val0 / 100), u => 'hPa' }
            };
            $ii++;
        } elsif (/\G010051/ && defined $val0) {
            if (exists $report{gpSurface}) {
                $report{ERROR} = { errorType => 'other', s => 'gpSurface' };
                last;
            }
            $report{SLP} = {
                s        => $s0,
                pressure => { v => sprintf('%.1f', $val0 / 100), u => 'hPa' }
            };
            $ii++;
        } elsif (/\G010061/) {
            if (exists $report{pressureChange}) {
                $report{pressureChange}{s} .= " $s0";
            } else {
                $report{pressureChange} = {
                    s             => $s0,
                    timeBeforeObs => { hours => 3 }
                };
            }
            if (defined $val0) {
               delete $report{pressureChange}{notAvailable};
               $report{pressureChange}{pressureChangeVal} =
                               { v => sprintf('%.1f', $val0 / 100), u => 'hPa' };
            } else {
                $report{pressureChange}{notAvailable} = undef;
            }
            $ii++;
        } elsif (exists $report{isSynop} && /\G010062/) {
            push @{$report{section3}}, { pressureChange => {
                    s             => $s0,
                    timeBeforeObs => { hours => 24 },
                    defined $val0
                      ? (pressureChangeVal => {
                              v => sprintf('%.1f', $val0 / 100),
                              u => 'hPa' })
                      : (notAvailable => undef)
            }};
            $ii++;
        } elsif (/\G010063/) {
            if (exists $report{pressureChange}) {
                $report{pressureChange}{s} .= " $s0";
            } else {
                $report{pressureChange} = {
                    s             => $s0,
                    notAvailable  => undef,
                    timeBeforeObs => { hours => 3 }
                };
            }
            if (defined $val0) {
                if ($val0 >= 0 && $val0 <= 8) {
                    $report{pressureChange}{pressureTendency} = $val0
                        if exists $report{pressureChange}{pressureChangeVal};
                } else {
                    push @{$report{ERROR}}, {
                        errorType => 'other',
                        s         => $s0
                    };
                }
            }
            $ii++;
        } elsif (/\G013023/) {
            push @{$report{section3}}, { precipitation => {
                    s => $s0,
                      !defined $val0 ? (notAvailable => undef)
                    : (timeBeforeObs => { hours => 24 },
                       $val0 == -0.1
                                ? (precipTraces => undef)
                                : (precipAmount => { v => $val0, u => 'MM' }))
            }};
            $ii++;
        } elsif (exists $report{isSynop} && /\G004024 013011/) {
            if (defined $val0 && defined $data->[$ii + 1]) {
                push @{$report{section3}}, { precipitation => {
                        s             => _mk_s($data, $desc, $ii, 2),
                        timeBeforeObs => { hours => 0-$val0 },
                        $data->[$ii + 1] == -0.1
                            ? (precipTraces => undef)
                            : (precipAmount => { v => $data->[$ii + 1],
                                                 u => 'MM' })
                }};
            } else {
                push @{$report{section3}}, { precipitation => {
                        s            => _mk_s($data, $desc, $ii, 2),
                        notAvailable => undef
                }};
            }
            $ii += 2;
        } elsif (/\G020001/ && !exists $report{visPrev}) {
            $report{visPrev} = {
                s => $s0,
                defined $val0 ? (distance => { v => $val0, u => 'M' })
                              : (notAvailable => undef)
            };
            $ii++;
        } elsif (/\G020010/ && !exists $report{totalCloudCover}) {
            if (defined $val0 && $val0 > 100 && $val0 != 113) {
                push @{$report{ERROR}}, {
                    errorType => 'other',
                    s         => $s0
                };
            } else {
                $report{totalCloudCover} = {
                    s => $s0,
                      !defined $val0 ? (notAvailable => undef)
                    : $val0 == 113   ? (skyObscured => undef)
                    :                  (oktas => sprintf '%.0f', $val0 / 12.5)
                };
            }
            $ii++;
        } elsif (   /\G008002 020011 020013 020012 020012 020012/
                 && !exists $report{cloudTypes})
        {
            $report{baseLowestCloud} = {
                s => _mk_s($data, $desc, $ii + 2, 1),
                defined $data->[$ii + 2]
                    ? (distance => { v => $data->[$ii + 2], u => 'M' })
                    : (notAvailable => undef)
            };

            $report{cloudTypes}{s} = _mk_s($data, $desc, $ii, 2) . ' ' .
                                     _mk_s($data, $desc, $ii + 3, 3);
            if (!defined $data->[$ii + 1] && !defined $data->[$ii + 3]) {
                $report{cloudTypes}{notAvailable} = undef;
            } elsif (defined $data->[$ii + 1] && $data->[$ii + 1] == 9) {
                $report{cloudTypes}{skyObscured} = undef;
            } else {
                if (defined $data->[$ii + 3] && $data->[$ii + 3] < 40) {
                    $report{cloudTypes}{cloudTypeLow} = $data->[$ii + 3] - 30;
                } else {
                    $report{cloudTypes}{cloudTypeLowNA} = undef;
                }
                if (defined $data->[$ii + 4] && $data->[$ii + 4] < 30) {
                    $report{cloudTypes}{cloudTypeMiddle} = $data->[$ii + 4] - 20;
                } else {
                    $report{cloudTypes}{cloudTypeMiddleNA} = undef;
                }
                if (defined $data->[$ii + 5] && $data->[$ii + 5] < 20) {
                    $report{cloudTypes}{cloudTypeHigh} = $data->[$ii + 5] - 10;
                } else {
                    $report{cloudTypes}{cloudTypeHighNA} = undef;
                }
                if (   defined $val0
                    && ($val0 == 7 || $val0 == 8)
                    && defined $data->[$ii + 1])
                {
                    $report{cloudTypes}{$val0 == 7 ? 'oktasLow' : 'oktasMiddle'}
                                                             = $data->[$ii + 1];
                }
            }

            $ii += 6;
        } elsif (   exists $report{isSynop}
                 && /\G008002 020011 020012 020013/
                 && defined $val0
                 && $val0 >= 1 && $val0 <= 3)
        {
            my $r;

            $r->{s} = _mk_s($data, $desc, $ii, 4);
            if (!defined $data->[$ii + 3]) {
                push @{$r->{sortedArr}}, { cloudBaseNotAvailable => undef };
            } else {
                push @{$r->{sortedArr}}, { cloudBase => {
                    v => $data->[$ii + 3],
                    u => 'M'
                }};
            }
            push @{$r->{sortedArr}},
                !defined $data->[$ii + 2] || $data->[$ii + 2] > 9
                ? { cloudTypeNotAvailable => undef }
                : { cloudType =>
                          qw(CI CC CS AC AS NS SC ST CU CB)[$data->[$ii + 2]] };
            push @{$r->{sortedArr}}, { cloudOktas => {
                  !defined $data->[$ii + 1] ? (oktasNotAvailable => undef)
                : $data->[$ii + 1] == 9     ? (skyObscured => undef)
                :                             (oktas => $data->[$ii + 1])
            }};
            push @{$report{section3}}, { cloudInfo => $r };
            $ii += 4;
        } elsif (   /\G(002002 )?008021 004025 011001 011002 008021((?: 004025 011043 011041)*)/
                 && !exists $report{sfcWind}
                 && defined $data->[$ii + ($1 ? 1 : 0)] && $data->[$ii + ($1 ? 1 : 0)] == 2
                 && !defined $data->[$ii + 4 + ($1 ? 1 : 0)])
        {
            $report{sfcWind}{s} = _mk_s($data, $desc, $ii, 5 + ($1 ? 1 : 0));
            $ii++
                if $1;

            if (defined $data->[$ii + 1]) {
                if ($data->[$ii + 1] > 0) {
                    push @{$report{ERROR}}, {
                        errorType => 'other',
                        s         => $report{sfcWind}{s}
                    };
                    $data->[$ii + 1] = -$data->[$ii + 1];
                }
                $report{sfcWind}{measurePeriod} =
                                        { v => 0-$data->[$ii + 1], u => 'MIN' };
            }

            if (!defined $data->[$ii + 2] && !defined $data->[$ii + 3]) {
                $report{sfcWind}{wind}{notAvailable} = undef;
            } elsif (   defined $data->[$ii + 2] && defined $data->[$ii + 3]
                     && $data->[$ii + 2] == 0    && $data->[$ii + 3] == 0)
            {
                $report{sfcWind}{wind}{isCalm} = undef;
            } else {
                if (defined $data->[$ii + 2]) {
                    $report{sfcWind}{wind}{dir}{v} = $data->[$ii + 2];
                } else {
                    $report{sfcWind}{wind}{dirNotAvailable} = undef;
                }
                if (defined $data->[$ii + 3]) {
                    $report{sfcWind}{wind}{speed} =
                                          { v => $data->[$ii + 3], u => 'MPS' };
                } else {
                    $report{sfcWind}{wind}{speedNotAvailable} = undef;
                }
            }
            $ii += 5;

            for (1 .. length($2) / 21) {
                if (   defined $data->[$ii]
                    && defined $data->[$ii + 1]
                    && defined $data->[$ii + 2])
                {
                    push @{$report{section3}}, { highestGust => {
                        s             => _mk_s($data, $desc, $ii, 3),
                        measurePeriod => { v => 0-$data->[$ii], u => 'MIN' },
                        wind          => {
                          defined $data->[$ii + 1] ? (dir => $data->[$ii + 1])
                                                   : (dirNotAvailable => undef),
                          defined $data->[$ii + 2]
                              ? (speed => { v => $data->[$ii + 2], u => 'MPS' })
                              : (speedNotAvailable => undef)
                        }
                    }};
                } else {
                    $report{sfcWind}{s} .= ' ' . _mk_s($data, $desc, $ii, 3);
                }
                $ii += 3;
            }
        } elsif (   /\G004025 011042/
                 && defined $val0
                 && defined $data->[$ii + 1])
        {
            push @{$report{section3}}, { highestMeanSpeed => {
                s             => _mk_s($data, $desc, $ii, 2),
                measurePeriod => { v => 10, u => 'MIN' },
                wind          => { speed => { v => $data->[$ii + 1], u=>'MPS'}},
                timeBeforeObs => { hours => -$val0 / 60 }
            }};
            $ii += 2;
        } elsif (   /\G007032 004024 004024 012111 004024 004024 012112/
                 && $val0 // 1.5 == 1.5)
        {
            push @{$report{section3}}, { tempMax => {
                s => _mk_s($data, $desc, $ii, 4),
                   defined $data->[$ii + 1]
                && defined $data->[$ii + 2]
                && defined $data->[$ii + 3]
                    ? (temp          => _K2C($data->[$ii + 3]),
                       timeBeforeObs => ($data->[$ii + 2] == 0)
                        ? { hours => 0-$data->[$ii + 1] }
                        : { hoursFrom => 0-$data->[$ii + 1],
                            hoursTill => 0-$data->[$ii + 2] })
                    : (notAvailable => undef)
            }};
            $ii += 4;
            push @{$report{section3}}, { tempMin => {
                s => _mk_s($data, $desc, $ii, 3),
                   defined $data->[$ii]
                && defined $data->[$ii + 1]
                && defined $data->[$ii + 2]
                    ? (temp          => _K2C($data->[$ii + 2]),
                       timeBeforeObs => ($data->[$ii + 1] == 0)
                        ? { hours => 0-$data->[$ii] }
                        : { hoursFrom => 0-$data->[$ii],
                            hoursTill => 0-$data->[$ii + 1] })
                    : (notAvailable => undef)
            }};
            $ii += 3;
        } elsif (/\G012113/) {
            push @{$report{section3}}, { tempMinGround => {
                s => $s0,
                defined $val0 ? (temp          => _K2C($val0),
                                 timeBeforeObs => { hours => 12 })
                              : (notAvailable => undef)
            }};
            $ii++;
        } elsif (   /\G020003 004024 020004 020005/
                 && !exists $report{weatherSynop})
        {
            if (defined $val0) {
                my $wxIndVal;

                $wxIndVal =   $val0 < 100  ? 1
                            : $val0 < 200  ? 7
                            : $val0 == 508 ? ($is_auto ? 5 : 2)
                            : $val0 == 509 ? ($is_auto ? 6 : 3)
                            :                0;
                $report{wxInd} = { s => $wxIndVal, wxIndVal => $wxIndVal }
                    if $wxIndVal;
            }

            $report{weatherSynop} = {
                s             => _mk_s($data, $desc, $ii, 4),
                timeBeforeObs => defined $data->[$ii + 1]
                    ? { hours => 0-$data->[$ii + 1] }
                    : { notAvailable => undef }
            };
            if (!defined $val0 || $val0 >= 300) {
                $report{weatherSynop}{weatherPresentNotAvailable} = undef;
            } elsif ($val0 < 100) {
                $report{weatherSynop}{weatherPresent} = $val0;
            } elsif ($val0 < 200) {
                $report{weatherSynop}{weatherPresentSimple} = $val0 - 100;
            } else {
                # must be processed by metaf2xml::parser
                push @{$report{weatherSynopAdd}}, {
                    s              => $s0,
                    weatherPresent => $val0 - 200,
                };
                $report{weatherSynop}{weatherPresentNotAvailable} = undef;
            }

            if (!defined $data->[$ii + 2] || $data->[$ii + 2] > 19) {
                $report{weatherSynop}{weatherPast1NotAvailable} = undef;
            } elsif ($data->[$ii + 2] < 10) {
                $report{weatherSynop}{weatherPast1} = $data->[$ii + 2];
            } else {
                $report{weatherSynop}{weatherPast1Simple} =
                                                          $data->[$ii + 2] - 10;
            }

            if (!defined $data->[$ii + 3] || $data->[$ii + 3] > 19) {
                $report{weatherSynop}{weatherPast2NotAvailable} = undef;
            } elsif ($data->[$ii + 3] < 10) {
                $report{weatherSynop}{weatherPast2} = $data->[$ii + 3];
            } else {
                $report{weatherSynop}{weatherPast2Simple} =
                                                          $data->[$ii + 3] - 10;
            }

            $ii += 4;
        } elsif (/\G004025 004025 020003/) {
            push @{$report{section3}}, { weatherSynopAmpl => {
                s              => _mk_s($data, $desc, $ii, 3),
                timeBeforeObs  =>      defined $val0
                                    && defined $data->[$ii + 1]
                                    && $data->[$ii + 1] == 0
                                  ? { hours => -$val0 / 60 }
                                  : { notAvailable => undef },
                weatherPresent => $data->[$ii + 2]
            }};
            $ii += 3;
        } elsif (/\G020003/ && defined $val0 && $val0 < 100) {
            push @{$report{section3}}, { weatherSynopAdd => {
                s              => $s0,
                weatherPresent => $val0
            }};
            $ii++;
        } elsif (/\G004024 014031/ && defined $val0) {
            # TODO: sunshinePeriod/@v=24 != sunshinePeriod/@v=p !?!
            _addRadiationSun \%report, {
                s => _mk_s($data, $desc, $ii, 2),
                defined $data->[$ii + 1]
                    ? (sunshine => { v => $data->[$ii + 1], u => 'MIN' })
                    : (sunshineNotAvailable => undef),
                sunshinePeriod => { v => 0-$val0, u => 'H' }
            };
            $ii += 2;
        } elsif (   /\G004024 014002 014004 014016 014028 014029 014030/
                 && defined $val0)
        {
            _addRadiationSun \%report, {
                s => _mk_s($data, $desc, $ii,
                           6 + (defined $data->[$ii + 6] ? 0 : 1)),
                defined $data->[$ii + 1]
                    ? (rad4DownwardLongWave => { radiationValue =>
                                 { v => $data->[$ii + 1] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 2]
                    ? (rad6ShortWave => { radiationValue =>
                                 { v => $data->[$ii + 2] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 3] && $data->[$ii + 3] >= 0
                    ? (rad0PosNet => { radiationValue =>
                                 { v => $data->[$ii + 3] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 3] && $data->[$ii + 3] < 0
                    ? (rad1NegNet => { radiationValue =>
                                { v => -$data->[$ii + 3] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 4]
                    ? (rad2GlobalSolar => { radiationValue =>
                                 { v => $data->[$ii + 4] / 1000, u => 'kJm2' }})
                    : (),
                defined $data->[$ii + 5]
                    ? (rad3DiffusedSolar => { radiationValue =>
                                 { v => $data->[$ii + 5] / 1000, u => 'kJm2' }})
                    : (),
                radiationPeriod      => { v => 0-$val0, u => 'H' },
                sunshinePeriod       => { v => 0-$val0, u => 'H' },
                sunshineNotAvailable => undef,
            };
            if (defined $data->[$ii + 6]) {
                push @{$report{section3}}, { radDirectSolar => {
                    s               => "$s0 " . _mk_s($data, $desc, $ii + 6, 1),
                    radiationPeriod => { v => 0-$val0, u => 'H' },
                    radiationValue  => { v => $data->[$ii + 6], u => 'kJm2' }
                }};
            }
            $ii += 7;
        } elsif (/\G002149/) {
            if (defined $val0) {
                $report{buoyType} = { s => $s0, buoyTypeInd => $val0 };
            } else {
                $report{buoyType} = { s => $s0, notAvailable => undef };
            }
            $ii++;
        } elsif (exists $report{isSynop} && /\G((?:007061 012130 )+)007061/) {
            for (1..length($1) / 14) {
                my $len = $_ == length($1) / 14 ? 3 : 2;

                if (defined $data->[$ii] && defined $data->[$ii + 1]) {
                    push @{$report{section5}}, { soilTemp => {
                        s => _mk_s($data, $desc, $ii, $len),
                        depth => { v => $data->[$ii], u => 'M' },
                        temp => _K2C $data->[$ii + 1]
                    }};
                } else {
                    $not_processed_s .= ' ' . _mk_s($data, $desc, $ii, $len);
                }
                $ii += $len;
            }
        } elsif (m{\G(?:031001|031000|001051|002036|002148|007031|007032|007033|008081|008082|025026|027004|028004|012064|001023)})
        {
            # these are ignored
            $ii++;
        } else {
            $is_processed = 0;
            $not_processed_s .= ' ' . $s0;
            $ii++;
        }

        if (($is_processed || $ii > $#$desc) && $not_processed_s ne '') {
            push @{$report{warning}}, {
                    warningType => 'notProcessed',
                    s           => substr $not_processed_s, 1
            };
            $not_processed_s = '';
        }
    }

    return %report;
}

=head1 SEE ALSO

L<metaf2xml::parser|metaf2xml::parser>(3pm),
L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2012-2015 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This file is part of metaf2xml.

metaf2xml is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

metaf2xml is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with metaf2xml; if not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA, or see
L<http://www.gnu.org/licenses/>.

=cut

1;
